###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""basic_modules defines basic VisTrails Modules that are used in most
pipelines."""
from __future__ import division

import vistrails.core.cache.hasher
from vistrails.core.debug import format_exception
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.vistrails_module import Module, new_module, \
    Converter, NotCacheable, ModuleError
from vistrails.core.modules.config import ConstantWidgetConfig, \
    QueryWidgetConfig, ParamExpWidgetConfig, ModuleSettings, IPort, OPort, \
    CIPort
import vistrails.core.system
from vistrails.core.utils import InstanceObject
from vistrails.core import debug

from abc import ABCMeta
from ast import literal_eval
from itertools import izip
import mimetypes
import os
import pickle
import re
import shutil
import zipfile
import urllib

try:
    import hashlib
    sha_hash = hashlib.sha1
except ImportError:
    import sha
    sha_hash = sha.new

###############################################################################

version = '2.1.1'
name = 'Basic Modules'
identifier = 'org.vistrails.vistrails.basic'
old_identifiers = ['edu.utah.sci.vistrails.basic']

constant_config_path = "vistrails.gui.modules.constant_configuration"
query_config_path = "vistrails.gui.modules.query_configuration"
paramexp_config_path = "vistrails.gui.modules.paramexplore"

def get_port_name(port):
    if hasattr(port, 'name'):
        return port.name
    else:
        return port[0]

class meta_add_value_ports(type):
    def __new__(cls, name, bases, dct):
        """This metaclass adds the 'value' input and output ports.
        """
        mod = type.__new__(cls, name, bases, dct)

        if '_input_ports' in mod.__dict__:
            input_ports = mod._input_ports
            if not any(get_port_name(port_info) == 'value'
                       for port_info in input_ports):
                mod._input_ports = [('value', mod)]
                mod._input_ports.extend(input_ports)
        else:
            mod._input_ports = [('value', mod)]

        if '_output_ports' in mod.__dict__:
            output_ports = mod._output_ports
            if not any(get_port_name(port_info) == 'value'
                       for port_info in output_ports):
                mod._output_ports = [('value', mod)]
                mod._output_ports.extend(output_ports)
        else:
            mod._output_ports = [('value', mod)]

        return mod

class Constant(Module):
    """Base class for all Modules that represent a constant value of
    some type.
    
    When implementing your own constant, You have to adhere to the
    following interface:

    Implement the following methods:
    
       translate_to_python(x): Given a string, translate_to_python
       must return a python value that will be the value seen by the
       execution modules.

       For example, translate_to_python called on a float parameter
       with value '3.15' will return float('3.15').
       
       translate_to_string(): Return a string representation of the
       current constant, which will eventually be passed to
       translate_to_python.

       validate(v): return True if given python value is a plausible
       value for the constant. It should be implemented such that
       validate(translate_to_python(x)) == True for all valid x

    A constant must also expose its default value, through the field
    default_value.

    There are fields you are not allowed to use in your constant classes.
    These are: 'id', 'interpreter', 'logging' and 'change_parameter'

    You can also define the constant's own GUI widget.
    See core/modules/constant_configuration.py for details.
    
    """
    _settings = ModuleSettings(abstract=True)
    _output_ports = [OPort("value_as_string", "String")]

    __metaclass__ = meta_add_value_ports

    def compute(self):
        """Constant.compute() only checks validity (and presence) of
        input value."""
        v = self.get_input("value")
        b = self.validate(v)
        if not b:
            raise ModuleError(self, "Internal Error: Constant failed validation")
        self.set_output("value", v)
        self.set_output("value_as_string", self.translate_to_string(v))

    def setValue(self, v):
        self.set_output("value", self.translate_to_python(v))
        self.upToDate = True

    @staticmethod
    def translate_to_string(v):
        return str(v)

    @staticmethod
    def get_widget_class():
        # return StandardConstantWidget
        return None

    @staticmethod
    def query_compute(value_a, value_b, query_method):
        if query_method == '==' or query_method is None:
            return (value_a == value_b)
        elif query_method == '!=':
            return (value_a != value_b)
        return False

def new_constant(name, py_conversion=None, default_value=None, validation=None,
                 widget_type=None,
                 str_conversion=None, base_class=Constant,
                 compute=None, query_compute=None):
    """new_constant(name: str, 
                    py_conversion: callable,
                    default_value: python_type,
                    validation: callable,
                    widget_type: (path, name) tuple or QWidget type,
                    str_conversion: callable,
                    base_class: class,
                    compute: callable,
                    query_compute: static callable) -> Module

    new_constant dynamically creates a new Module derived from
    Constant with given py_conversion and str_conversion functions, a
    corresponding python type and a widget type. py_conversion is a
    python callable that takes a string and returns a python value of
    the type that the class should hold. str_conversion does the reverse.

    This is the quickest way to create new Constant Modules."""

    d = {}

    if py_conversion is not None:
        d["translate_to_python"] = py_conversion
    elif base_class == Constant:
        raise ValueError("Must specify translate_to_python for constant")
    if validation is not None:
        d["validate"] = validation
    elif base_class == Constant:
        raise ValueError("Must specify validation for constant")
    if default_value is not None:
        d["default_value"] = default_value

    if str_conversion is not None:
        d['translate_to_string'] = str_conversion
    if compute is not None:
        d['compute'] = compute
    if query_compute is not None:
        d['query_compute'] = query_compute
    if widget_type is not None:
        @staticmethod
        def get_widget_class():
            return widget_type
        d['get_widget_class'] = get_widget_class            

    m = new_module(base_class, name, d)
    m._input_ports = [('value', m)]
    m._output_ports = [('value', m)]
    return m

class Boolean(Constant):
    _settings = ModuleSettings(
            constant_widget='%s:BooleanWidget' % constant_config_path)
    default_value = False

    @staticmethod
    def translate_to_python(x):
        s = x.upper()
        if s == 'TRUE':
            return True
        if s == 'FALSE':
            return False
        raise ValueError('Boolean from String in VisTrails should be either '
                         '"true" or "false", got "%s" instead' % x)

    @staticmethod
    def validate(x):
        return isinstance(x, bool)

class Float(Constant):
    _settings = ModuleSettings(constant_widgets=[
        QueryWidgetConfig('%s:NumericQueryWidget' % query_config_path),
        ParamExpWidgetConfig('%s:FloatExploreWidget' % paramexp_config_path)])
    default_value = 0.0

    @staticmethod
    def translate_to_python(x):
        return float(x)

    @staticmethod
    def validate(x):
        return isinstance(x, (int, long, float))

    @staticmethod
    def query_compute(value_a, value_b, query_method):
        value_a = float(value_a)
        value_b = float(value_b)
        if query_method == '==' or query_method is None:
            return (value_a == value_b)
        elif query_method == '<':
            return (value_a < value_b)
        elif query_method == '>':
            return (value_a > value_b)
        elif query_method == '<=':
            return (value_a <= value_b)
        elif query_method == '>=':
            return (value_a >= value_b)

class Integer(Float):
    _settings = ModuleSettings(constant_widgets=[
        QueryWidgetConfig('%s:NumericQueryWidget' % query_config_path),
        ParamExpWidgetConfig('%s:IntegerExploreWidget' % paramexp_config_path)])
    default_value = 0

    @staticmethod
    def translate_to_python(x):
        if x.startswith('0x'):
            return int(x, 16)
        else:
            return int(x)

    @staticmethod
    def validate(x):
        return isinstance(x, (int, long))

class String(Constant):
    _settings = ModuleSettings(
            configure_widget="vistrails.gui.modules.string_configure:TextConfigurationWidget",
            constant_widgets=[
                 ConstantWidgetConfig('%s:MultiLineStringWidget' % constant_config_path,
                                      widget_type='multiline'),
                 QueryWidgetConfig('%s:StringQueryWidget' % query_config_path)])
    _output_ports = [OPort("value_as_string", "String", optional=True)]
    default_value = ""

    @staticmethod
    def translate_to_python(x):
        assert isinstance(x, (str, unicode))
        return str(x)

    @staticmethod
    def validate(x):
        return isinstance(x, str)

    @staticmethod
    def query_compute(value_a, value_b, query_method):
        if query_method == '*[]*' or query_method is None:
            return (value_b in value_a)
        elif query_method == '==':
            return (value_a == value_b)
        elif query_method == '=~':
            try:
                m = re.match(value_b, value_a)
                if m is not None:
                    return (m.end() ==len(value_a))
            except re.error:
                pass
        return False

##############################################################################

# Rich display for IPython
try:
    from IPython import display
except ImportError:
    display = None

class PathObject(object):
    def __init__(self, name):
        self.name = name
        self._ipython_repr = None

    def __repr__(self):
        return "PathObject(%r)" % self.name
    __str__ = __repr__

    def __getattr__(self, name):
        if name.startswith('_repr_') and name.endswith('_'):
            if self._ipython_repr is None:
                filetype, encoding = mimetypes.guess_type(self.name)
                if not filetype:
                    self._ipython_repr = False
                elif filetype.startswith('image/'):
                    self._ipython_repr = display.Image(filename=self.name)
                else:
                    self._ipython_repr = False
            elif self._ipython_repr is not False:
                return getattr(self._ipython_repr, name)
        raise AttributeError

class Path(Constant):
    _settings = ModuleSettings(constant_widget=("%s:PathChooserWidget" % \
                                                constant_config_path))
    _input_ports = [IPort("value", "Path"),
                    IPort("name", "String", optional=True)]
    _output_ports = [OPort("value", "Path")]

    @staticmethod
    def translate_to_python(x):
        return PathObject(x)

    @staticmethod
    def translate_to_string(x):
        return str(x.name)

    @staticmethod
    def validate(v):
        return isinstance(v, PathObject)

    def get_name(self):
        n = None
        if self.has_input("value"):
            n = self.get_input("value").name
        if n is None:
            self.check_input("name")
            n = self.get_input("name")
        return n

    def set_results(self, n):
        self.set_output("value", PathObject(n))
        self.set_output("value_as_string", n)

    def compute(self):
        n = self.get_name()
        self.set_results(n)

Path.default_value = PathObject('')

def path_parameter_hasher(p):
    def get_mtime(path):
        t = int(os.path.getmtime(path))
        if os.path.isdir(path):
            for subpath in os.listdir(path):
                subpath = os.path.join(path, subpath)
                if os.path.isdir(subpath):
                    t = max(t, get_mtime(subpath))
        return t

    h = vistrails.core.cache.hasher.Hasher.parameter_signature(p)
    try:
        # FIXME: This will break with aliases - I don't really care that much
        t = get_mtime(p.strValue)
    except OSError:
        return h
    hasher = sha_hash()
    hasher.update(h)
    hasher.update(str(t))
    return hasher.digest()

class File(Path):
    """File is a VisTrails Module that represents a file stored on a
    file system local to the machine where VisTrails is running."""

    _settings = ModuleSettings(constant_signature=path_parameter_hasher,
                               constant_widget=("%s:FileChooserWidget" % \
                                                constant_config_path))
    _input_ports = [IPort("value", "File"),
                    IPort("create_file", "Boolean", optional=True)]
    _output_ports = [OPort("value", "File"),
                     OPort("local_filename", "String", optional=True)]

    def compute(self):
        n = self.get_name()
        if (self.has_input("create_file") and self.get_input("create_file")):
            vistrails.core.system.touch(n)
        if not os.path.isfile(n):
            raise ModuleError(self, 'File %r does not exist' % n)
        self.set_results(n)
        self.set_output("local_filename", n)

class Directory(Path):

    _settings = ModuleSettings(constant_signature=path_parameter_hasher,
                               constant_widget=("%s:DirectoryChooserWidget" % \
                                                constant_config_path))
    _input_ports = [IPort("value", "Directory"),
                    IPort("create_directory", "Boolean", optional=True)]
    _output_ports = [OPort("value", "Directory"),
                     OPort("itemList", "List")]

    def compute(self):
        n = self.get_name()
        if (self.has_input("create_directory") and 
                self.get_input("create_directory")):
            try:
                vistrails.core.system.mkdir(n)
            except Exception, e:
                raise ModuleError(self, 'mkdir: %s' % format_exception(e))
        if not os.path.isdir(n):
            raise ModuleError(self, 'Directory "%s" does not exist' % n)
        self.set_results(n)

        dir_list = os.listdir(n)
        output_list = []
        for item in dir_list:
            full_path = os.path.join(n, item)
            output_list.append(PathObject(full_path))
        self.set_output('itemList', output_list)

##############################################################################

class OutputPath(Path):
    _settings = ModuleSettings(constant_widget=("%s:OutputPathChooserWidget" % \
                                                constant_config_path))
    _input_ports = [IPort("value", "OutputPath")]
    _output_ports = [OPort("value", "OutputPath")]

    def get_name(self):
        n = None
        if self.has_input("value"):
            n = self.get_input("value").name
        if n is None:
            self.check_input("name")
            n = self.get_input("name")
        return n

    def set_results(self, n):
        self.set_output("value", PathObject(n))
        self.set_output("value_as_string", n)

    def compute(self):
        n = self.get_name()
        self.set_results(n)

class FileSink(NotCacheable, Module):
    """FileSink takes a file and writes it to a user-specified
    location in the file system.  The file is stored at location
    specified by the outputPath.  The overwrite flag allows users to
    specify whether an existing path should be overwritten."""

    _input_ports = [IPort("file", File),
                    IPort("outputPath", OutputPath),
                    IPort("overwrite", Boolean, optional=True, 
                          default=True),
                    IPort("publishFile", Boolean, optional=True)]
    
    def compute(self):
        input_file = self.get_input("file")
        output_path = self.get_input("outputPath")
        full_path = output_path.name

        if os.path.isfile(full_path):
            if self.get_input('overwrite'):
                try:
                    os.remove(full_path)
                except OSError, e:
                    msg = ('Could not delete existing path "%s" '
                           '(overwrite on)' % full_path)
                    raise ModuleError(self, msg)
            else:
                raise ModuleError(self,
                                  "Could not copy file to '%s': file already "
                                  "exists")

        try:
            vistrails.core.system.link_or_copy(input_file.name, full_path)
        except OSError, e:
            msg = "Could not create file '%s': %s" % (full_path, e)
            raise ModuleError(self, msg)

        if (self.has_input("publishFile") and
            self.get_input("publishFile") or 
            not self.has_input("publishFile")):
            if self.moduleInfo.has_key('extra_info'):
                if self.moduleInfo['extra_info'].has_key('pathDumpCells'):
                    folder = self.moduleInfo['extra_info']['pathDumpCells']
                    base_fname = os.path.basename(full_path)
                    (base_fname, file_extension) = os.path.splitext(base_fname)
                    base_fname = os.path.join(folder, base_fname)
                    # make a unique filename
                    filename = base_fname + file_extension
                    counter = 2
                    while os.path.exists(filename):
                        filename = base_fname + "_%d%s" % (counter,
                                                           file_extension)
                        counter += 1
                    try:
                        vistrails.core.system.link_or_copy(input_file.name, filename)
                    except OSError, e:
                        msg = "Could not publish file '%s' \n   on  '%s':" % (
                                full_path, filename)
                        # I am not sure whether we should raise an error
                        # I will just print a warning for now (Emanuele)
                        debug.warning("%s" % msg, e)

class DirectorySink(NotCacheable, Module):
    """DirectorySink takes a directory and writes it to a
    user-specified location in the file system.  The directory is
    stored at location specified by the outputPath.  The overwrite
    flag allows users to specify whether an existing path should be
    overwritten."""

    _input_ports = [IPort("dir", Directory),
                    IPort("outputPath", OutputPath),
                    IPort("overwrite", Boolean, optional=True, default="True")]

    def compute(self):
        input_dir = self.get_input("dir")
        output_path = self.get_input("outputPath")
        full_path = output_path.name

        if os.path.exists(full_path):
            if self.get_input("overwrite"):
                try:
                    if os.path.isfile(full_path):
                        os.remove(full_path)
                    else:
                        shutil.rmtree(full_path)
                except OSError, e:
                    msg = ('Could not delete existing path "%s" '
                           '(overwrite on)' % full_path)
                    raise ModuleError(
                            self,
                            '%s\n%s' % (msg, format_exception(e)))
            else:
                msg = ('Could not write to existing path "%s" '
                       '(overwrite off)' % full_path)
                raise ModuleError(self, msg)
            
        try:
            shutil.copytree(input_dir.name, full_path)
        except OSError, e:
            msg = 'Could not copy path from "%s" to "%s"' % \
                (input_dir.name, full_path)
            raise ModuleError(self, '%s\n%s' % (msg, format_exception(e)))

##############################################################################

class WriteFile(Converter):
    """Writes a String to a temporary File.
    """
    _input_ports = [IPort('in_value', String),
                    IPort('suffix', String, optional=True, default=""),
                    IPort('encoding', String, optional=True)]
    _output_ports = [OPort('out_value', File)]

    def compute(self):
        contents = self.get_input('in_value')
        suffix = self.force_get_input('suffix', '')
        result = self.interpreter.filePool.create_file(suffix=suffix)
        if self.has_input('encoding'):
            contents = contents.decode('utf-8') # VisTrails uses UTF-8
                                                # internally (I hope)
            contents = contents.encode(self.get_input('encoding'))
        with open(result.name, 'wb') as fp:
            fp.write(contents)
        self.set_output('out_value', result)

class ReadFile(Converter):
    """Reads a File to a String.
    """
    _input_ports = [IPort('in_value', File),
                    IPort('encoding', String, optional=True)]
    _output_ports = [OPort('out_value', String)]

    def compute(self):
        filename = self.get_input('in_value').name
        with open(filename, 'rb') as fp:
            contents = fp.read()
        if self.has_input('encoding'):
            contents = contents.decode(self.get_input('encoding'))
            contents = contents.encode('utf-8') # VisTrails uses UTF-8
                                                # internally (for now)
        self.set_output('out_value', contents)

##############################################################################

class Color(Constant):
    # We set the value of a color object to be an InstanceObject that
    # contains a tuple because a tuple would be interpreted as a
    # type(tuple) which messes with the interpreter

    _settings = ModuleSettings(constant_widgets=[
        '%s:ColorWidget' % constant_config_path, 
        ConstantWidgetConfig('%s:ColorEnumWidget' % \
                             constant_config_path, 
                             widget_type='enum'),
        QueryWidgetConfig('%s:ColorQueryWidget' % \
                          query_config_path),
        ParamExpWidgetConfig('%s:RGBExploreWidget' % \
                             paramexp_config_path,
                             widget_type='rgb'),
        ParamExpWidgetConfig('%s:HSVExploreWidget' % \
                             paramexp_config_path,
                             widget_type='hsv')])
    _input_ports = [IPort("value", "Color")]
    _output_ports = [OPort("value", "Color")]

    default_value = InstanceObject(tuple=(1,1,1))

    @staticmethod
    def translate_to_python(x):
        return InstanceObject(
            tuple=tuple([float(a) for a in x.split(',')]))

    @staticmethod
    def translate_to_string(v):
        return ','.join('%f' % c for c in v.tuple)

    @staticmethod
    def validate(x):
        return isinstance(x, InstanceObject) and hasattr(x, 'tuple')

    @staticmethod
    def to_string(r, g, b):
        return "%s,%s,%s" % (r,g,b)        

    @staticmethod
    def query_compute(value_a, value_b, query_method):
        # SOURCE: http://www.easyrgb.com/index.php?X=MATH
        def rgb_to_xyz(r, g, b):
            # r,g,b \in [0,1]

            if r > 0.04045: 
                r = ( ( r + 0.055 ) / 1.055 ) ** 2.4
            else:
                r = r / 12.92
            if g > 0.04045:
                g = ( ( g + 0.055 ) / 1.055 ) ** 2.4
            else:
                g = g / 12.92
            if b > 0.04045: 
                b = ( ( b + 0.055 ) / 1.055 ) ** 2.4
            else:
                b = b / 12.92

            r *= 100
            g *= 100
            b *= 100

            # Observer. = 2 deg, Illuminant = D65
            x = r * 0.4124 + g * 0.3576 + b * 0.1805
            y = r * 0.2126 + g * 0.7152 + b * 0.0722
            z = r * 0.0193 + g * 0.1192 + b * 0.9505
            return (x,y,z)

        def xyz_to_cielab(x,y,z):
            # Observer= 2 deg, Illuminant= D65
            ref_x, ref_y, ref_z = (95.047, 100.000, 108.883)
            x /= ref_x
            y /= ref_y
            z /= ref_z

            if x > 0.008856:
                x = x ** ( 1/3.0 )
            else:                    
                x = ( 7.787 * x ) + ( 16 / 116.0 )
            if y > 0.008856:
                y = y ** ( 1/3.0 )
            else:
                y = ( 7.787 * y ) + ( 16 / 116.0 )
            if z > 0.008856: 
                z = z ** ( 1/3.0 )
            else:
                z = ( 7.787 * z ) + ( 16 / 116.0 )

            L = ( 116 * y ) - 16
            a = 500 * ( x - y )
            b = 200 * ( y - z )
            return (L, a, b)

        def rgb_to_cielab(r,g,b):
            return xyz_to_cielab(*rgb_to_xyz(r,g,b))
        
        value_a_rgb = (float(a) for a in value_a.split(','))
        value_b_rgb = (float(b) for b in value_b.split(','))
        value_a_lab = rgb_to_cielab(*value_a_rgb)
        value_b_lab = rgb_to_cielab(*value_b_rgb)
        
        # cie76 difference
        diff = sum((v_1 - v_2) ** 2 
                   for v_1, v_2 in izip(value_a_lab, value_b_lab)) ** (0.5)

        # print "CIE 76 DIFFERENCE:", diff
        if query_method is None:
            query_method = '2.3'
        return diff < float(query_method)

##############################################################################

class StandardOutput(NotCacheable, Module):
    """StandardOutput is a VisTrails Module that simply prints the
    value connected on its port to standard output. It is intended
    mostly as a debugging device."""

    _input_ports = [IPort("value", 'Variant')]

    def compute(self):
        v = self.get_input("value")
        if isinstance(v, PathObject):
            try:
                fp = open(v.name, 'rb')
            except IOError:
                print v
            else:
                try:
                    CHUNKSIZE = 2048
                    chunk = fp.read(CHUNKSIZE)
                    if chunk:
                        sys.stdout.write(chunk)
                        while len(chunk) == CHUNKSIZE:
                            chunk = fp.read(CHUNKSIZE)
                            if chunk:
                                sys.stdout.write(chunk)
                        sys.stdout.write('\n')
                finally:
                    fp.close()
        else:
            print v

##############################################################################

# Tuple will be reasonably magic right now. We'll integrate it better
# with vistrails later.
# TODO: Check Tuple class, test, integrate.
class Tuple(Module):
    """Tuple represents a tuple of values. Tuple might not be well
    integrated with the rest of VisTrails, so don't use it unless
    you know what you're doing."""

    _settings = ModuleSettings(configure_widget=
        "vistrails.gui.modules.tuple_configuration:TupleConfigurationWidget")

    def __init__(self):
        Module.__init__(self)
        self.input_ports_order = []
        self.values = tuple()

    def transfer_attrs(self, module):
        Module.transfer_attrs(self, module)
        self.input_ports_order = [p.name for p in module.input_port_specs]

    def compute(self):
        values = tuple([self.get_input(p)
                        for p in self.input_ports_order])
        self.values = values
        self.set_output("value", values)

class Untuple(Module):
    """Untuple takes a tuple and returns the individual values.  It
    reverses the actions of Tuple.

    """

    _settings = ModuleSettings(configure_widget=
        "vistrails.gui.modules.tuple_configuration:UntupleConfigurationWidget")

    def __init__(self):
        Module.__init__(self)
        self.output_ports_order = []

    def transfer_attrs(self, module):
        Module.transfer_attrs(self, module)
        self.output_ports_order = [p.name for p in module.output_port_specs]
        # output_ports are reversed for display purposes...
        self.output_ports_order.reverse()

    def compute(self):
        if self.has_input("tuple"):
            tuple = self.get_input("tuple")
            values = tuple.values
        else:
            values = self.get_input("value")
        for p, value in izip(self.output_ports_order, values):
            self.set_output(p, value)

##############################################################################

class ConcatenateString(Module):
    """ConcatenateString takes many strings as input and produces the
    concatenation as output. Useful for constructing filenames, for
    example.

    This class will probably be replaced with a better API in the
    future."""

    fieldCount = 4
    _input_ports = [IPort("str%d" % i, "String")
                    for i in xrange(1, 1 + fieldCount)]
    _output_ports = [OPort("value", "String")]

    def compute(self):
        result = "".join(self.force_get_input('str%d' % i, '')
                         for i in xrange(1, 1 + self.fieldCount))
        self.set_output('value', result)

##############################################################################

class Not(Module):
    """Not inverts a Boolean.
    """
    _input_ports = [IPort('input', 'Boolean')]
    _output_ports = [OPort('value', 'Boolean')]

    def compute(self):
        value = self.get_input('input')
        self.set_output('value', not value)

##############################################################################

# List

# If numpy is available, we consider numpy arrays to be lists as well
class ListType(object):
    __metaclass__ = ABCMeta

ListType.register(list)
try:
    import numpy
except ImportError:
    numpy = None
else:
    ListType.register(numpy.ndarray)

class List(Constant):
    _settings = ModuleSettings(configure_widget=
        "vistrails.gui.modules.list_configuration:ListConfigurationWidget")
    _input_ports = [IPort("value", "List"),
                    IPort("head", "Variant", depth=1),
                    IPort("tail", "List")]
    _output_ports = [OPort("value", "List")]

    default_value = []

    def __init__(self):
        Constant.__init__(self)
        self.input_ports_order = []

    def transfer_attrs(self, module):
        Module.transfer_attrs(self, module)
        self.input_ports_order = [p.name for p in module.input_port_specs]

    @staticmethod
    def validate(x):
        return isinstance(x, ListType)

    @staticmethod
    def translate_to_python(v):
        return literal_eval(v)

    @staticmethod
    def translate_to_string(v, dims=None):
        if dims is None:
            if numpy is not None and isinstance(v, numpy.ndarray):
                dims = v.ndim
            else:
                dims = 1
        if dims == 1:
            return '[%s]' % ', '.join(repr(c)
                                      for c in v)
        else:
            return '[%s]' % ', '.join(List.translate_to_string(c, dims-1)
                                      for c in v)

    def compute(self):
        head, middle, items, tail = [], [], [], []
        got_value = False

        if self.has_input('value'):
            # run the regular compute here
            Constant.compute(self)
            middle = self.outputPorts['value']
            got_value = True
        if self.has_input('head'):
            head = self.get_input('head')
            got_value = True
        if self.input_ports_order:
            items = [self.get_input(p)
                     for p in self.input_ports_order]
            got_value = True
        if self.has_input('tail'):
            tail = self.get_input('tail')
            got_value = True

        if not got_value:
            self.get_input('value')
        self.set_output('value', head + middle + items + tail)

##############################################################################
# Dictionary

class Dictionary(Constant):
    default_value = {}
    _input_ports = [CIPort("addPair", "Module, Module"),
                    IPort("addPairs", "List")]

    @staticmethod
    def translate_to_python(v):
        return literal_eval(v)

    @staticmethod
    def validate(x):
        return isinstance(x, dict)

    def compute(self):
        d = {}
        if self.has_input('value'):
            Constant.compute(self)
            d.update(self.outputPorts['value'])
        if self.has_input('addPair'):
            pairs_list = self.get_input_list('addPair')
            d.update(pairs_list)
        if self.has_input('addPairs'):
            d.update(self.get_input('addPairs'))

        self.set_output("value", d)

##############################################################################

# TODO: Null should be a subclass of Constant?
class Null(Module):
    """Null is the class of None values."""
    _settings = ModuleSettings(hide_descriptor=True)

    def compute(self):
        self.set_output("value", None)

##############################################################################

class Unpickle(Module):
    """Unpickles a string.
    """
    _settings = ModuleSettings(hide_descriptor=True)
    _input_ports = [IPort('input', 'String')]
    _output_ports = [OPort('result', 'Variant')]

    def compute(self):
        value = self.get_input('input')
        self.set_output('result', pickle.loads(value))

##############################################################################

class CodeRunnerMixin(object):
    def __init__(self):
        self.output_ports_order = []
        super(CodeRunnerMixin, self).__init__()

    def transfer_attrs(self, module):
        Module.transfer_attrs(self, module)
        self.output_ports_order = [p.name for p in module.output_port_specs]
        # output_ports are reversed for display purposes...
        self.output_ports_order.reverse()

    def run_code(self, code_str,
                 use_input=False,
                 use_output=False):
        """run_code runs a piece of code as a VisTrails module.
        use_input and use_output control whether to use the inputport
        and output port dictionary as local variables inside the
        execution."""
        import vistrails.core.packagemanager
        def fail(msg):
            raise ModuleError(self, msg)
        def cache_this():
            self.is_cacheable = lambda *args, **kwargs: True
        locals_ = locals()
        if use_input:
            for k in self.inputPorts:
                locals_[k] = self.get_input(k)
        if use_output:
            for output_portname in self.output_ports_order:
                locals_[output_portname] = None
        _m = vistrails.core.packagemanager.get_package_manager()
        reg = get_module_registry()
        locals_.update({'fail': fail,
                        'package_manager': _m,
                        'cache_this': cache_this,
                        'registry': reg,
                        'self': self})
        if 'source' in locals_:
            del locals_['source']
        # Python 2.6 needs code to end with newline
        exec code_str + '\n' in locals_, locals_
        if use_output:
            for k in self.output_ports_order:
                if locals_.get(k) is not None:
                    self.set_output(k, locals_[k])

##############################################################################

class PythonSource(CodeRunnerMixin, NotCacheable, Module):
    """PythonSource is a Module that executes an arbitrary piece of
    Python code.

    It is especially useful for one-off pieces of 'glue' in a
    pipeline.

    If you want a PythonSource execution to fail, call
    fail(error_message).

    If you want a PythonSource execution to be cached, call
    cache_this().
    """
    _settings = ModuleSettings(
        configure_widget=("vistrails.gui.modules.python_source_configure:"
                             "PythonSourceConfigurationWidget"))
    _input_ports = [IPort('source', 'String', optional=True, default="")]
    _output_pors = [OPort('self', 'Module')]

    def compute(self):
        s = urllib.unquote(str(self.get_input('source')))
        self.run_code(s, use_input=True, use_output=True)

##############################################################################

def zip_extract_file(archive, filename_in_archive, output_filename):
    z = zipfile.ZipFile(archive)
    try:
        fileinfo = z.getinfo(filename_in_archive) # Might raise KeyError
        output_dirname, output_filename = os.path.split(output_filename)
        fileinfo.filename = output_filename
        z.extract(fileinfo, output_dirname)
    finally:
        z.close()


def zip_extract_all_files(archive, output_path):
    z = zipfile.ZipFile(archive)
    try:
        z.extractall(output_path)
    finally:
        z.close()


class Unzip(Module):
    """Unzip extracts a file from a ZIP archive."""
    _input_ports = [IPort('archive_file', 'File'),
                    IPort('filename_in_archive', 'String')]
    _output_ports = [OPort('file', 'File')]

    def compute(self):
        self.check_input("archive_file")
        self.check_input("filename_in_archive")
        filename_in_archive = self.get_input("filename_in_archive")
        archive_file = self.get_input("archive_file")
        if not os.path.isfile(archive_file.name):
            raise ModuleError(self, "archive file does not exist")
        suffix = self.interpreter.filePool.guess_suffix(filename_in_archive)
        output = self.interpreter.filePool.create_file(suffix=suffix)
        zip_extract_file(archive_file.name,
                         filename_in_archive,
                         output.name)
        self.set_output("file", output)


class UnzipDirectory(Module):
    """UnzipDirectory extracts every file from a ZIP archive."""
    _input_ports = [IPort('archive_file', 'File')]
    _output_ports = [OPort('directory', 'Directory')]

    def compute(self):
        self.check_input("archive_file")
        archive_file = self.get_input("archive_file")
        if not os.path.isfile(archive_file.name):
            raise ModuleError(self, "archive file does not exist")
        output = self.interpreter.filePool.create_directory()
        zip_extract_all_files(archive_file.name,
                              output.name)
        self.set_output("directory", output)

##############################################################################

class Round(Converter):
    """Turns a Float into an Integer.
    """
    _settings = ModuleSettings(hide_descriptor=True)
    _input_ports = [IPort('in_value', 'Float'),
                    IPort('floor', 'Boolean', optional=True, default="True")]
    _output_ports = [OPort('out_value', 'Integer')]

    def compute(self):
        fl = self.get_input('in_value')
        floor = self.get_input('floor')
        if floor:
            integ = int(fl)         # just strip the decimals
        else:
            integ = int(fl + 0.5)   # nearest
        self.set_output('out_value', integ)


class TupleToList(Converter):
    """Turns a Tuple into a List.
    """
    _settings = ModuleSettings(hide_descriptor=True)
    _input_ports = [IPort('in_value', 'Variant')]
    _output_ports = [OPort('out_value', 'List')]
    
    @classmethod
    def can_convert(cls, sub_descs, super_descs):
        if len(sub_descs) <= 1:
            return False
        reg = get_module_registry()
        return super_descs == [reg.get_descriptor(List)]

    def compute(self):
        tu = self.get_input('in_value')
        if not isinstance(tu, tuple):
            raise ModuleError(self, "Input is not a tuple")
        self.set_output('out_value', list(tu))

##############################################################################
    
class Variant(Module):
    """
    Variant is tracked internally for outputing a variant type on
    output port. For input port, Module type should be used
    
    """
    _settings = ModuleSettings(abstract=True)

##############################################################################

class Generator(object):
    """
    Used to keep track of list iteration, it will execute a module once for
    each input in the list/generator.
    """
    _settings = ModuleSettings(abstract=True)

    generators = []
    def __init__(self, size=None, module=None, generator=None, port=None,
                 accumulated=False):
        self.module = module
        self.generator = generator
        self.port = port
        self.size = size
        self.accumulated = accumulated
        if generator and module not in Generator.generators:
            # add to global list of generators
            # they will be topologically ordered
            module.generator = generator
            Generator.generators.append(module)
            
    def next(self):
        """ return next value - the generator """
        value = self.module.get_output(self.port)
        if isinstance(value, Generator):
            value = value.all()
        return value
    
    def all(self):
        """ exhausts next() for Streams
        
        """
        items = []
        item = self.next()
        while item is not None:
            items.append(item)
            item = self.next()
        return items

    @staticmethod
    def stream():
        """ executes all generators until inputs are exhausted
            this makes sure branching and multiple sinks are executed correctly

        """
        result = True
        if not Generator.generators:
            return
        while result is not None:
            for g in Generator.generators:
                result = g.next()
        Generator.generators = []

##############################################################################

class Assert(Module):
    """
    Assert is a simple module that conditionally stops the execution.
    """
    _input_ports = [IPort('condition', 'Boolean')]

    def compute(self):
        condition = self.get_input('condition')
        if not condition:
            raise ModuleError(self, "Assert: condition is False",
                              abort=True)


class AssertEqual(Module):
    """
    AssertEqual works like Assert but compares two values.

    It is provided for convenience.
    """

    _input_ports = [IPort('value1', 'Variant'),
                    IPort('value2', 'Variant')]

    def compute(self):
        values = (self.get_input('value1'),
                  self.get_input('value2'))
        if values[0] != values[1]:
            reprs = tuple(repr(v) for v in values)
            reprs = tuple('%s...' % v[:17] if len(v) > 20 else v
                          for v in reprs)
            raise ModuleError(self, "AssertEqual: values are different: "
                                    "%r, %r" % reprs,
                              abort=True)

##############################################################################

class StringFormat(Module):
    """
    Builds a string from objects using Python's str.format().
    """
    _settings = ModuleSettings(configure_widget=
        'vistrails.gui.modules.stringformat_configuration:'
            'StringFormatConfigurationWidget')
    _input_ports = [IPort('format', String)]
    _output_ports = [OPort('value', String)]

    @staticmethod
    def list_placeholders(fmt):
        placeholders = set()
        nb = 0
        i = 0
        n = len(fmt)
        while i < n:
            if fmt[i] == '{':
                i += 1
                if fmt[i] == '{': # KeyError:
                    i += 1
                    continue
                e = fmt.index('}', i) # KeyError
                f = e
                for c in (':', '!', '[', '.'):
                    c = fmt.find(c, i)
                    if c != -1:
                        f = min(f, c)
                if i == f:
                    nb += 1
                else:
                    arg = fmt[i:f]
                    try:
                        arg = int(arg)
                    except ValueError:
                        placeholders.add(arg)
                    else:
                        nb = max(nb, arg + 1)
                i = e
            i += 1
        return nb, placeholders

    def compute(self):
        fmt = self.get_input('format')
        args, kwargs = StringFormat.list_placeholders(fmt)
        f_args = [self.get_input('_%d' % n)
                  for n in xrange(args)]
        f_kwargs = dict((n, self.get_input(n))
                        for n in kwargs)
        self.set_output('value', fmt.format(*f_args, **f_kwargs))

##############################################################################

def init_constant(m):
    reg = get_module_registry()

    reg.add_module(m)
    reg.add_input_port(m, "value", m)
    reg.add_output_port(m, "value", m)

_modules = [Module, Converter, Constant, Boolean, Float, Integer, String, List,
            Path, File, Directory, OutputPath,
            FileSink, DirectorySink, WriteFile, ReadFile, StandardOutput,
            Tuple, Untuple, ConcatenateString, Not, Dictionary, Null, Variant,
            Unpickle, PythonSource, Unzip, UnzipDirectory, Color,
            Round, TupleToList, Assert, AssertEqual, StringFormat]

def initialize(*args, **kwargs):
    # initialize the sub_module modules, too
    import vistrails.core.modules.sub_module
    import vistrails.core.modules.output_modules
    _modules.extend(vistrails.core.modules.sub_module._modules)
    _modules.extend(vistrails.core.modules.output_modules._modules)


def handle_module_upgrade_request(controller, module_id, pipeline):
    from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler
    reg = get_module_registry()

    def outputName_remap(old_conn, new_module):
        ops = []
        old_src_module = pipeline.modules[old_conn.source.moduleId]
        op_desc = reg.get_descriptor(OutputPath)
        new_x = (old_src_module.location.x + new_module.location.x) / 2.0
        new_y = (old_src_module.location.y + new_module.location.y) / 2.0
        op_module = \
            controller.create_module_from_descriptor(op_desc, new_x, new_y)
        ops.append(('add', op_module))
        create_new_connection = UpgradeWorkflowHandler.create_new_connection
        new_conn_1 = create_new_connection(controller,
                                           old_src_module,
                                           old_conn.source,
                                           op_module,
                                           "name")
        ops.append(('add', new_conn_1))
        new_conn_2 = create_new_connection(controller,
                                           op_module,
                                           "value",
                                           new_module,
                                           "outputPath")
        ops.append(('add', new_conn_2))
        return ops

    module_remap = {'FileSink':
    [(None, '1.6', None,
                          {'dst_port_remap':
                               {'overrideFile': 'overwrite',
                                'outputName': outputName_remap},
                           'function_remap':
                               {'overrideFile': 'overwrite',
                                'outputName': 'outputPath'}})],
                    'GetItemsFromDirectory':
                        [(None, '1.6', 'Directory',
                          {'dst_port_remap':
                               {'dir': 'value'},
                           'src_port_remap':
                               {'itemlist': 'itemList'},
                           })],
                    'InputPort':
                        [(None, '1.6', None,
                          {'dst_port_remap': {'old_name': None}})],
                    'OutputPort':
                        [(None, '1.6', None,
                          {'dst_port_remap': {'old_name': None}})],
                    'PythonSource':
                        [(None, '1.6', None, {})],
                    'Tuple':
                        [(None, '2.1.1', None, {})],
                    'StandardOutput':
                        [(None, '2.1.1', None, {})],
                    'List':
                        [(None, '2.1.1', None, {})],
                    'AssertEqual':
                        [(None, '2.1.1', None, {})],
                    'Converter':
                        [(None, '2.1.1', None, {})],
                    }

    return UpgradeWorkflowHandler.remap_module(controller, module_id, pipeline,
                                               module_remap)

###############################################################################

class NewConstant(Constant):
    """
    A new Constant module to be used inside the FoldWithModule module.
    """
    def setValue(self, v):
        self.set_output("value", v)
        self.upToDate = True

def create_constant(value):
    """
    Creates a NewConstant module, to be used for the ModuleConnector.
    """
    constant = NewConstant()
    constant.setValue(value)
    return constant

def get_module(value, signature=None):
    """
    Creates a module for value, in order to do the type checking.
    """
    if isinstance(value, Constant):
        return type(value)
    elif isinstance(value, bool):
        return Boolean
    elif isinstance(value, str):
        return String
    elif isinstance(value, int):
        return Integer
    elif isinstance(value, float):
        return Float
    if isinstance(value, list):
        return List
    elif isinstance(value, tuple):
        # Variant supports signatures of any length
        if signature is None or \
           (len(signature) == 1 and signature[0][0] == Variant):
            return (Variant,)*len(value)
        v_modules = ()
        for element in xrange(len(value)):
            v_modules += (get_module(value[element], signature[element]),)
        if None in v_modules: # Identification failed
            return None
        return v_modules
    else: # pragma: no cover
        debug.warning("Could not identify the type of the list element.")
        debug.warning("Type checking is not going to be done inside "
                      "iterated module.")
        return None

###############################################################################

import sys
import unittest

class TestConcatenateString(unittest.TestCase):
    @staticmethod
    def concatenate(**kwargs):
        from vistrails.tests.utils import execute, intercept_result
        with intercept_result(ConcatenateString, 'value') as results:
            errors = execute([
                    ('ConcatenateString', 'org.vistrails.vistrails.basic', [
                        (name, [('String', value)])
                        for name, value in kwargs.iteritems()
                    ]),
                ])
            if errors:
                return None
        return results

    def test_concatenate(self):
        """Concatenates strings"""
        self.assertEqual(self.concatenate(
                str1="hello ", str2="world"),
                ["hello world"])
        self.assertEqual(self.concatenate(
                str3="hello world"),
                ["hello world"])
        self.assertEqual(self.concatenate(
                str2="hello ", str4="world"),
                ["hello world"])
        self.assertEqual(self.concatenate(
                str1="hello", str3=" ", str4="world"),
                ["hello world"])

    def test_empty(self):
        """Runs ConcatenateString with no input"""
        self.assertEqual(self.concatenate(), [""])


class TestNot(unittest.TestCase):
    def run_pipeline(self, functions):
        from vistrails.tests.utils import execute, intercept_result
        with intercept_result(Not, 'value') as results:
            errors = execute([
                    ('Not', 'org.vistrails.vistrails.basic',
                     functions),
                ])
        return errors, results

    def test_true(self):
        errors, results = self.run_pipeline([
                ('input', [('Boolean', 'True')])])
        self.assertFalse(errors)
        self.assertEqual(len(results), 1)
        self.assertIs(results[0], False)

    def test_false(self):
        errors, results = self.run_pipeline([
                ('input', [('Boolean', 'False')])])
        self.assertFalse(errors)
        self.assertEqual(len(results), 1)
        self.assertIs(results[0], True)

    def test_notset(self):
        errors, results = self.run_pipeline([])
        self.assertTrue(errors)


class TestList(unittest.TestCase):
    @staticmethod
    def build_list(value=None, head=None, tail=None):
        from vistrails.tests.utils import execute, intercept_result
        with intercept_result(List, 'value') as results:
            functions = []
            def add(n, v, t):
                if v is not None:
                    for e in v:
                        functions.append(
                                (n, [(t, e)])
                            )
            add('value', value, 'List')
            add('head', head, 'String')
            add('tail', tail, 'List')

            errors = execute([
                    ('List', 'org.vistrails.vistrails.basic', functions),
                ])
            if errors:
                return None
        # List is a Constant, so the interpreter will set the result 'value'
        # from the 'value' input port automatically
        # Ignore these first results
        return results[-1]

    def test_simple(self):
        """Tests the default ports of the List module"""
        self.assertEqual(self.build_list(
                value=['["a", "b", "c"]']),
                ["a", "b", "c"])
        self.assertEqual(self.build_list(
                head=["d"],
                value=['["a", "b", "c"]']),
                ["d", "a", "b", "c"])
        self.assertEqual(self.build_list(
                head=["d"],
                value=['["a", "b", "c"]'],
                tail=['["e", "f"]']),
                ["d", "a", "b", "c", "e", "f"])
        self.assertEqual(self.build_list(
                value=['[]'],
                tail=['[]']),
                [])

    def test_multiple(self):
        """Tests setting multiple values on a port"""
        # Multiple values on 'head'
        self.assertEqual(self.build_list(
                head=["a", "b"]),
                ["a", "b"])
        self.assertEqual(self.build_list(
                head=["a", "b"],
                value=['["c", "d"]']),
                ["a", "b", "c", "d"])

        # Multiple values on 'value'
        res = self.build_list(value=['["a", "b"]', '["c", "d"]'])
        # Connections of List type are merged
        self.assertEqual(res, ["a", "b", "c", "d"])

    def test_items(self):
        """Tests the multiple 'itemN' ports"""
        from vistrails.tests.utils import execute, intercept_result
        def list_with_items(nb_items, **kwargs):
            with intercept_result(List, 'value') as results:
                errors = execute([
                        ('List', 'org.vistrails.vistrails.basic', [
                            (k, [('String', v)])
                            for k, v in kwargs.iteritems()
                        ]),
                    ],
                    add_port_specs=[
                        (0, 'input', 'item%d' % i,
                         '(org.vistrails.vistrails.basic:Module)')
                        for i in xrange(nb_items)
                    ])
                if errors:
                    return None
            return results[-1]

        self.assertEqual(
                list_with_items(2, head="one", item0="two", item1="three"),
                ["one", "two", "three"])

        # All 'itemN' ports have to be set
        self.assertIsNone(
                list_with_items(3, head="one", item0="two", item2="three"))


class TestPythonSource(unittest.TestCase):
    def test_simple(self):
        """A simple PythonSource returning a string"""
        import urllib2
        from vistrails.tests.utils import execute, intercept_result
        source = 'customout = "nb is %d" % customin'
        source = urllib2.quote(source)
        with intercept_result(PythonSource, 'customout') as results:
            self.assertFalse(execute([
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', source)]),
                        ('customin', [('Integer', '42')])
                    ]),
                    ('String', 'org.vistrails.vistrails.basic', []),
                ],
                [
                    (0, 'customout', 1, 'value'),
                ],
                add_port_specs=[
                    (0, 'input', 'customin',
                     'org.vistrails.vistrails.basic:Integer'),
                    (0, 'output', 'customout',
                     'org.vistrails.vistrails.basic:String'),
                ]))
        self.assertEqual(results[-1], "nb is 42")


class TestNumericConversions(unittest.TestCase):
    def test_full(self):
        from vistrails.tests.utils import execute, intercept_result
        with intercept_result(Round, 'out_value') as results:
            self.assertFalse(execute([
                    ('Integer', 'org.vistrails.vistrails.basic', [
                        ('value', [('Integer', '5')])
                    ]),
                    ('Float', 'org.vistrails.vistrails.basic', []),
                    ('PythonCalc', 'org.vistrails.vistrails.pythoncalc', [
                        ('value2', [('Float', '2.7')]),
                        ('op', [('String', '+')]),
                    ]),
                    ('Round', 'org.vistrails.vistrails.basic', [
                        ('floor', [('Boolean', 'True')]),
                    ]),
                ],
                [
                    (0, 'value', 1, 'value'),
                    (1, 'value', 2, 'value1'),
                    (2, 'value', 3, 'in_value'),
                ]))
        self.assertEqual(results, [7])


class TestUnzip(unittest.TestCase):
    def test_unzip_file(self):
        from vistrails.tests.utils import execute, intercept_result
        from vistrails.core.system import vistrails_root_directory
        zipfile = os.path.join(vistrails_root_directory(),
                               'tests', 'resources',
                               'test_archive.zip')
        with intercept_result(Unzip, 'file') as outfiles:
            self.assertFalse(execute([
                    ('Unzip', 'org.vistrails.vistrails.basic', [
                        ('archive_file', [('File', zipfile)]),
                        ('filename_in_archive', [('String', 'file1.txt')]),
                    ]),
                ]))
        self.assertEqual(len(outfiles), 1)
        with open(outfiles[0].name, 'rb') as outfile:
            self.assertEqual(outfile.read(), "some random\ncontent")

    def test_unzip_all(self):
        from vistrails.tests.utils import execute, intercept_result
        from vistrails.core.system import vistrails_root_directory
        zipfile = os.path.join(vistrails_root_directory(),
                               'tests', 'resources',
                               'test_archive.zip')
        with intercept_result(UnzipDirectory, 'directory') as outdir:
            self.assertFalse(execute([
                    ('UnzipDirectory', 'org.vistrails.vistrails.basic', [
                        ('archive_file', [('File', zipfile)]),
                    ]),
                ]))
        self.assertEqual(len(outdir), 1)

        self.assertEqual(
                [(d, f) for p, d, f in os.walk(outdir[0].name)],
                [(['subdir'], ['file1.txt']),
                 ([], ['file2.txt'])])


from vistrails.core.configuration import get_vistrails_configuration

class TestTypechecking(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        conf = get_vistrails_configuration()
        cls.error_all = conf.showConnectionErrors
        cls.error_variant = conf.showVariantErrors

    @classmethod
    def tearDownClass(cls):
        conf = get_vistrails_configuration()
        conf.showConnectionErrors = cls.error_all
        conf.showVariantErrors = cls.error_variant

    @staticmethod
    def set_settings(error_all, error_variant):
        conf = get_vistrails_configuration()
        conf.showConnectionErrors = error_all
        conf.showVariantErrors = error_variant

    def run_test_pipeline(self, result, expected_results, *args, **kwargs):
        from vistrails.tests.utils import execute, intercept_result
        for error_all, error_variant, expected in expected_results:
            self.set_settings(error_all, error_variant)
            with intercept_result(*result) as results:
                error = execute(*args, **kwargs)
            if not expected:
                self.assertTrue(error)
            else:
                self.assertFalse(error)
                self.assertEqual(results, expected)

    def test_basic(self):
        import urllib2
        # Base case: no typing error
        # This should succeed in every case
        self.run_test_pipeline(
            (PythonSource, 'r'),
            [(False, False, ["test"]),
             (True, True, ["test"])],
            [
                ('PythonSource', 'org.vistrails.vistrails.basic', [
                    ('source', [('String', urllib2.quote('o = "test"'))]),
                ]),
                ('PythonSource', 'org.vistrails.vistrails.basic', [
                    ('source', [('String', urllib2.quote('r = i'))])
                ]),
            ],
            [
                (0, 'o', 1, 'i'),
            ],
            add_port_specs=[
                (0, 'output', 'o',
                 'org.vistrails.vistrails.basic:String'),
                (1, 'input', 'i',
                 'org.vistrails.vistrails.basic:String'),
                (1, 'output', 'r',
                 'org.vistrails.vistrails.basic:String')
            ])

    def test_fake(self):
        import urllib2
        # A module is lying, declaring a String but returning an int
        # This should fail with showConnectionErrors=True (not the
        # default)
        self.run_test_pipeline(
            (PythonSource, 'r'),
            [(False, False, [42]),
             (False, True, [42]),
             (True, True, False)],
            [
                ('PythonSource', 'org.vistrails.vistrails.basic', [
                    ('source', [('String', urllib2.quote('o = 42'))]),
                ]),
                ('PythonSource', 'org.vistrails.vistrails.basic', [
                    ('source', [('String', urllib2.quote('r = i'))])
                ]),
            ],
            [
                (0, 'o', 1, 'i'),
            ],
            add_port_specs=[
                (0, 'output', 'o',
                 'org.vistrails.vistrails.basic:String'),
                (1, 'input', 'i',
                 'org.vistrails.vistrails.basic:String'),
                (1, 'output', 'r',
                 'org.vistrails.vistrails.basic:String')
            ])

    def test_inputport(self):
        import urllib2
        # This test uses an InputPort module, whose output port should not be
        # considered a Variant port (although it is)
        self.run_test_pipeline(
            (PythonSource, 'r'),
            [(False, False, [42]),
             (False, True, [42]),
             (True, True, [42])],
            [
                ('InputPort', 'org.vistrails.vistrails.basic', [
                    ('ExternalPipe', [('Integer', '42')]),
                ]),
                ('PythonSource', 'org.vistrails.vistrails.basic', [
                    ('source', [('String', urllib2.quote('r = i'))])
                ]),
            ],
            [
                (0, 'InternalPipe', 1, 'i'),
            ],
            add_port_specs=[
                (1, 'input', 'i',
                 'org.vistrails.vistrails.basic:String'),
                (1, 'output', 'r',
                 'org.vistrails.vistrails.basic:String'),
            ])


class TestStringFormat(unittest.TestCase):
    def test_list_placeholders(self):
        fmt = 'a {} b}} {c!s} {{d e}} {}f'
        self.assertEqual(StringFormat.list_placeholders(fmt),
                         (2, set(['c'])))

    def run_format(self, fmt, expected, **kwargs):
        from vistrails.tests.utils import execute, intercept_result
        functions = [('format', [('String', fmt)])]
        functions.extend((n, [(t, v)])
                         for n, (t, v) in kwargs.iteritems())
        with intercept_result(StringFormat, 'value') as results:
            self.assertFalse(execute([
                    ('StringFormat', 'org.vistrails.vistrails.basic',
                     functions),
                ],
                add_port_specs=[
                    (0, 'input', n, t)
                    for n, (t, v) in kwargs.iteritems()
                ]))
        self.assertEqual(results, [expected])

    def test_format(self):
        self.run_format('{{ {a} }} b {c!s}', '{ 42 } b 12',
                        a=('Integer', '42'),
                        c=('Integer', '12'))

    # Python 2.6 doesn't support {}
    @unittest.skipIf(sys.version_info < (2, 7), "No {} support on 2.6")
    def test_format_27(self):
        self.run_format('{} {}', 'a b',
                        _0=('String', 'a'), _1=('String', 'b'))
        self.run_format('{{ {a} {} {b!s}', '{ 42 b 12',
                        a=('Integer', '42'), _0=('String', 'b'),
                        b=('Integer', '12'))
        self.run_format('{} {} {!r}{ponc} {:.2f}', "hello dear 'world'! 1.33",
                        _0=('String', 'hello'), _1=('String', 'dear'),
                        _2=('String', 'world'), _3=('Float', '1.333333333'),
                        ponc=('String', '!'))


class TestConstantMetaclass(unittest.TestCase):
    def test_meta(self):
        """Tests the __metaclass__ for Constant.
        """
        mod1_in = [('value', 'basic:String'), IPort('other', 'basic:Float')]
        mod1_out = [('someport', 'basic:Integer')]
        class Mod1(Constant):
            _input_ports = mod1_in
            _output_ports = mod1_out
        self.assertEqual(Mod1._input_ports, mod1_in)
        self.assertEqual(Mod1._output_ports, [('value', Mod1)] + mod1_out)

        mod2_in = [('another', 'basic:String')]
        class Mod2(Mod1):
            _input_ports = mod2_in
        self.assertEqual(Mod2._input_ports, [('value', Mod2)] + mod2_in)
        self.assertEqual(Mod2._output_ports, [('value', Mod2)])

        class Mod3(Mod1):
            _output_ports = []
        self.assertEqual(Mod3._input_ports, [('value', Mod3)])
        self.assertEqual(Mod3._output_ports, [('value', Mod3)])



from __future__ import division

from vistrails.core.bundles.pyimport import py_import
from vistrails.core.modules.utils import make_modules_dict
from vistrails.core.packagemanager import get_package_manager
from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler

try:
    py_import('numpy', {
            'pip': 'numpy',
            'linux-debian': 'python-numpy',
            'linux-ubuntu': 'python-numpy',
            'linux-fedora': 'numpy'})
except ImportError: # pragma: no cover
    pass

from .common import _modules as common_modules, TableOutput
from .convert import _modules as convert_modules
from .operations import _modules as operation_modules
from .read import _modules as read_modules
from .write import _modules as write_modules


_modules = [common_modules,
            convert_modules,
            operation_modules,
            read_modules,
            write_modules]

if get_package_manager().has_package( # pragma: no branch
        'org.vistrails.vistrails.spreadsheet'):
    from .viewer import _modules as viewer_modules, TableToSpreadsheetMode
    _modules.append(viewer_modules)
    TableOutput.register_output_mode(TableToSpreadsheetMode)

_modules = make_modules_dict(*_modules)


def handle_module_upgrade_request(controller, module_id, pipeline):
    def add_keyname(fname, module):
        new_function = controller.create_function(module,
                                                  "key_name",
                                                  ["_key"])
        return [('add', new_function, 'module', module.id)]

    module_remap = {
            'read|csv|CSVFile': [
                (None, '0.1.1', 'read|CSVFile', {
                    'src_port_remap': {
                        'self': 'value'},
                })
            ],
            'read|numpy|NumPyArray': [
                (None, '0.1.1', 'read|NumPyArray', {
                    'src_port_remap': {
                        'self': 'value'},
                })
            ],
            'read|CSVFile': [
                ('0.1.1', '0.1.2', None, {
                    'src_port_remap': {
                        'self': 'value'},
                }),
                ('0.1.3', '0.1.5', None, {})
            ],
            'read|NumPyArray': [
                ('0.1.1', '0.1.2', None, {
                    'src_port_remap': {
                        'self': 'value'},
                })
            ],
            'read|ExcelSpreadsheet': [
                ('0.1.1', '0.1.2', None, {
                    'src_port_remap': {
                        'self': 'value'},
                }),
                ('0.1.3', '0.1.4', None, {})
            ],
            'read|JSONFile': [
                (None, '0.1.5', 'read|JSONObject', {
                    'function_remap': {
                        None: add_keyname},
                })
            ],
        }

    try:
        from vistrails.packages.spreadsheet.init import upgrade_cell_to_output
    except ImportError:
        pass
    else:
        module_remap = upgrade_cell_to_output(
                module_remap, module_id, pipeline,
                'TableCell', 'TableOutput',
                '0.1.6', 'table')

    return UpgradeWorkflowHandler.remap_module(controller,
                                               module_id,
                                               pipeline,
                                               module_remap)



#!/usr/bin/env python

# In this example we show the use of the
# vtkBandedPolyDataContourFilter.  This filter creates separate,
# constant colored bands for a range of scalar values. Each band is
# bounded by two scalar values, and the cell data lying within the
# value has the same cell scalar value.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()


# The lookup table is similar to that used by maps. Two hues are used:
# a brown for land, and a blue for water. The value of the hue is
# changed to give the effect of elevation.
Scale = 5
lutWater = vtk.vtkLookupTable()
lutWater.SetNumberOfColors(10)
lutWater.SetHueRange(0.58, 0.58)
lutWater.SetSaturationRange(0.5, 0.1)
lutWater.SetValueRange(0.5, 1.0)
lutWater.Build()
lutLand = vtk.vtkLookupTable()
lutLand.SetNumberOfColors(10)
lutLand.SetHueRange(0.1, 0.1)
lutLand.SetSaturationRange(0.4, 0.1)
lutLand.SetValueRange(0.55, 0.9)
lutLand.Build()


# The DEM reader reads data and creates an output image.
demModel = vtk.vtkDEMReader()
demModel.SetFileName(VTK_DATA_ROOT + "/Data/SainteHelens.dem")
demModel.Update()

# We shrink the terrain data down a bit to yield better performance for 
# this example.
shrinkFactor = 4
shrink = vtk.vtkImageShrink3D()
shrink.SetShrinkFactors(shrinkFactor, shrinkFactor, 1)
shrink.SetInputConnection(demModel.GetOutputPort())
shrink.AveragingOn()

# Convert the image into polygons.
geom = vtk.vtkImageDataGeometryFilter()
geom.SetInputConnection(shrink.GetOutputPort())

# Warp the polygons based on elevation.
warp = vtk.vtkWarpScalar()
warp.SetInputConnection(geom.GetOutputPort())
warp.SetNormal(0, 0, 1)
warp.UseNormalOn()
warp.SetScaleFactor(Scale)

# Create the contour bands.
bcf = vtk.vtkBandedPolyDataContourFilter()
bcf.SetInput(warp.GetPolyDataOutput())
bcf.GenerateValues(15, demModel.GetOutput().GetScalarRange())
bcf.SetScalarModeToIndex()
bcf.GenerateContourEdgesOn()

# Compute normals to give a better look.
normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(bcf.GetOutputPort())
normals.SetFeatureAngle(60)
normals.ConsistencyOff()
normals.SplittingOff()

demMapper = vtk.vtkPolyDataMapper()
demMapper.SetInputConnection(normals.GetOutputPort())
demMapper.SetScalarRange(0, 10)
demMapper.SetLookupTable(lutLand)
demMapper.SetScalarModeToUseCellData()

demActor = vtk.vtkLODActor()
demActor.SetMapper(demMapper)

## Create contour edges
edgeMapper = vtk.vtkPolyDataMapper()
edgeMapper.SetInput(bcf.GetContourEdgesOutput())
edgeMapper.SetResolveCoincidentTopologyToPolygonOffset()
edgeActor = vtk.vtkActor()
edgeActor.SetMapper(edgeMapper)
edgeActor.GetProperty().SetColor(0, 0, 0)

## Test clipping
# Create the contour bands.
bcf2 = vtk.vtkBandedPolyDataContourFilter()
bcf2.SetInput(warp.GetPolyDataOutput())
bcf2.ClippingOn()
bcf2.GenerateValues(10, 1000, 2000)
bcf2.SetScalarModeToValue()

# Compute normals to give a better look.
normals2 = vtk.vtkPolyDataNormals()
normals2.SetInputConnection(bcf2.GetOutputPort())
normals2.SetFeatureAngle(60)
normals2.ConsistencyOff()
normals2.SplittingOff()

lut = vtk.vtkLookupTable()
lut.SetNumberOfColors(10)
demMapper2 = vtk.vtkPolyDataMapper()
demMapper2.SetInputConnection(normals2.GetOutputPort())
demMapper2.SetScalarRange(demModel.GetOutput().GetScalarRange())
demMapper2.SetLookupTable(lut)
demMapper2.SetScalarModeToUseCellData()

demActor2 = vtk.vtkLODActor()
demActor2.SetMapper(demMapper2)
demActor2.AddPosition(0, 15000, 0)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(demActor)
ren.AddActor(demActor2)
ren.AddActor(edgeActor)

ren.SetBackground(.4, .4, .4)
renWin.SetSize(375, 200)

cam = vtk.vtkCamera()
cam.SetPosition(-17438.8, 2410.62, 25470.8)
cam.SetFocalPoint(3985.35, 11930.6, 5922.14)
cam.SetViewUp(0, 0, 1)
ren.SetActiveCamera(cam)
ren.ResetCamera()
cam.Zoom(2)

iren.Initialize()
iren.SetDesiredUpdateRate(1)

def CheckAbort(obj, event):
    foo = renWin.GetEventPending()
    if foo != 0:
        renWin.SetAbortRender(1)

renWin.AddObserver("AbortCheckEvent", CheckAbort)
renWin.Render()

renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
import os
import re
import uuid

from vistrails.core.modules.basic_modules import Path
from vistrails.gui.common_widgets import QSearchBox
from vistrails.gui.modules.constant_configuration import ConstantWidgetMixin
from vistrails.gui.modules.module_configure import StandardModuleConfigurationWidget

from db_utils import DatabaseAccessSingleton
from identifiers import identifier as persistence_pkg
import repo

class IntegerWrapper(object):
    def __init__(self, idx):
        self.idx = idx

class PersistentRefModelSingleton(object):
    def __new__(self, *args, **kw):
        if PersistentRefModel._instance is None:
            obj = PersistentRefModel(*args, **kw)
            PersistentRefModel._instance = obj
        return PersistentRefModel._instance

class PersistentRefModel(QtCore.QAbstractItemModel):

    _instance = None

    # 2013-09-03 18:57:53.133000
    _DATE_FORMAT = re.compile(r'^(?P<y>[0-9]{4})-'
                              r'(?P<m>[0-9]{2})-'
                              r'(?P<d>[0-9]{2}) '
                              r'(?P<H>[0-9]{2}):'
                              r'(?P<M>[0-9]{2}):'
                              r'(?P<S>[0-9]{2}).'
                              r'(?P<ms>[0-9]+)$')

    cols = {0: "name",
            1: "type",
            2: "tags",
            3: "user",
            4: "date_created",
            5: "date_modified",
            6: "id",
            7: "version",
            8: "content_hash",
            9: "signature"}
    idxs = dict((v,k) for (k,v) in cols.iteritems())
    headers = {"id": "ID",
               "name": "Name",
               "tags": "Tags",
               "user": "User",
               "date_created": "Date Created",
               "date_modified": "Date Modified",
               "content_hash": "Content Hash",
               "version": "Version",
               "signature": "Signature",
               "type": "Type"}

    def __init__(self, parent=None):
        QtCore.QAbstractItemModel.__init__(self, parent)

        self.db_access = DatabaseAccessSingleton()
        self.db_access.set_model(self)
        rows = self.db_access.read_database(
            [c[1] for c in sorted(self.cols.iteritems())])
        rows = map(self.fix_dates, rows)

        self.id_lists = {}
        for ref in rows:
            if ref[self.idxs['id']] not in self.id_lists:
                self.id_lists[ref[self.idxs['id']]] = []
            self.id_lists[ref[self.idxs['id']]].append(ref)

        self.id_lists_keys = self.id_lists.keys()

        self.integer_wrappers = {}

    @staticmethod
    def fix_dates(row):
        row = list(row)
        for c in ('date_created', 'date_modified'):
            c = PersistentRefModel.idxs[c]
            m = PersistentRefModel._DATE_FORMAT.match(row[c])
            if m is not None:
                row[c] = '{y}-{m}-{d} {H}:{M}'.format(**m.groupdict())
        return tuple(row)

    def rowCount(self, parent=QtCore.QModelIndex()):
        if not parent.isValid():
            # print 'rowCount:', len(self.id_lists_keys)
            return len(self.id_lists_keys)
        # print 'parentValid rowCount:', \
        #     len(self.id_lists[self.id_lists_keys[parent.row()]])
        return len(self.id_lists[self.id_lists_keys[parent.row()]])
 
    def columnCount(self, parent=QtCore.QModelIndex()):
        # print 'columnCount:', len(self.headers)
        return len(self.headers)
    
    def hasChildren(self, parent=QtCore.QModelIndex()):
        # print 'hasChildren:'
        if not parent.isValid():
            # print '  True'
            return True
        else:
            # print '  PARENT:', parent.row(), parent.column(), \
            #     parent.internalPointer()
            if not parent.parent().isValid():
                # print '  TRUE'
                return True
        # print '  False'
        return False

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role != QtCore.Qt.DisplayRole:
            return None
        if section in self.cols:
            return self.headers[self.cols[section]]
        return None
    
    def data(self, index, role):
        if not index.isValid() or role != QtCore.Qt.DisplayRole:
            return None
        # if index.parent().isValid():
        #     print 'data', index.row(), index.column(), index.parent().row()
        if index.parent().isValid():
            # have a child node
            # print 'got child node!', index.parent().row()
            id_list = self.id_lists[self.id_lists_keys[index.parent().row()]]
            data = id_list[index.row()]
        else:
            # have a parent node    
            id_list = self.id_lists[self.id_lists_keys[index.row()]]
            data = id_list[0]
            # want to have the earliest created date and latest modified date
            if index.column() == self.idxs['date_created']:
                dates = [l[index.column()] for l in id_list]
                return min(dates)
            if index.column() == self.idxs['date_modified']:
                dates = [l[index.column()] for l in id_list]
                return max(dates)
            if index.column() == self.idxs['version'] or \
                    index.column() == self.idxs['signature'] or \
                    index.column() == self.idxs['content_hash'] or \
                    index.column() == self.idxs['user']:
                return None

        if index.column() < len(data):
            return data[index.column()]
        return None
    
    def parent(self, index):
        # print 'calling parent() method'
        if index.isValid():
            # print '  index is valid', index.row(), index.column()
            if index.internalPointer():
                parent_item = index.internalPointer().idx
                # print '  parent_item:', parent_item
                return self.createIndex(parent_item, 0, None)
#             else:
#                 # print '  internalPointer is not valid'
#         else:
#             # print 'index not valid for parent call'
        return QtCore.QModelIndex()

    def index(self, row, column, parent):
        # print 'index:', row, column
        if not parent.isValid():
            if len(self.id_lists_keys) > row:
                # print '  no parent item'
                return self.createIndex(row, column, None)
        else:
            # print '  **** parent_item', row, column, parent.row()
            if len(self.id_lists[self.id_lists_keys[parent.row()]]) > row:
                # print '  ++++ creating index'
                # !!! internalPointer is a weakref in PyQt !!!
                if parent.row() not in self.integer_wrappers:
                    integer_wrapper = IntegerWrapper(parent.row())
                    self.integer_wrappers[parent.row()] = integer_wrapper
                else:
                    integer_wrapper = self.integer_wrappers[parent.row()]
                # print '  ---- created parent wrapper'
                return self.createIndex(row, column, integer_wrapper)
        return QtCore.QModelIndex()

    def sort(self, column, order=QtCore.Qt.AscendingOrder):
        self.emit(QtCore.SIGNAL('layoutAboutToBeChanged()'))
        if column == -1:
            return
        self.id_lists_keys.sort(key=lambda x: self.id_lists[x][0][column], 
                                reverse=(order==QtCore.Qt.AscendingOrder))
        self.emit(QtCore.SIGNAL('layoutChanged()'))

    def find_row(self, id, version=None):
        if id in self.id_lists:
            i = self.id_lists_keys.index(id)
            if version is not None:
                for j, data in enumerate(self.id_lists[id]):
                    if data[self.idxs['version']] == version:
                        if i not in self.integer_wrappers:
                            integer_wrapper = IntegerWrapper(i)
                            self.integer_wrappers[i] = integer_wrapper
                        else:
                            integer_wrapper = self.integer_wrappers[i]
                        return self.createIndex(j, 0, integer_wrapper)
            return self.createIndex(i, 0, 0)
#             for i, id in enumerate(self.id_lists_keys):
            
#             if version is not None:
#                 for data in enumerate(self.id_lists[id]):
#                     if
#         for i, data in enumerate(self.id_lists):
#             if data[self.idxs['id']] == id and \
#                     (not version or data[self.idxs['version']] == version):
#                 return self.createIndex(i, 0, 0)
        return QtCore.QModelIndex()

    def add_data(self, value_dict):
        id = value_dict['id']
        value_list = []
        for _, c in sorted(self.cols.iteritems()):
            if c in value_dict:
                value_list.append(str(value_dict[c]))
            else:
                value_list.append(None)
        if id not in self.id_lists:
            self.id_lists[id] = []
            self.id_lists_keys.append(id)
        self.id_lists[id].append(self.fix_dates(value_list))
        self.reset()

    def remove_data(self, where_dict):
        id = where_dict['id']
        version = where_dict.get('version', None)
        if version is not None:
            for idx, value_tuple in enumerate(self.id_lists[id]):
                if value_tuple[self.idxs['version']] == version:
                    del self.id_lists[id][idx]
                    break
        else:
            path_type = self.id_lists[id][0][self.idxs['type']]
            for idx, key_id in enumerate(self.id_lists_keys):
                if key_id == id:
                    del self.id_lists_keys[idx]
                    break
            del self.id_lists[id]
        self.reset()

class PersistentRefView(QtGui.QTreeView):
    def __init__(self, path_type=None, parent=None):
        QtGui.QTreeView.__init__(self, parent)
        self.my_model = PersistentRefModelSingleton()
        print 'my_model:', id(self.my_model)
        proxy_model = QtGui.QSortFilterProxyModel(self)
        proxy_model.setSourceModel(self.my_model)
        proxy_model.setFilterKeyColumn(-1)
        self.setModel(proxy_model)
        self.set_visibility(path_type)

        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.setSortingEnabled(True)
        self.current_id = None
        self.current_version = None

        for i in xrange(self.my_model.columnCount()):
            self.resizeColumnToContents(i)

    def set_visibility(self, path_type=None):
        if path_type == "blob":
            self.header().hideSection(self.my_model.idxs["type"])
            for i, key in enumerate(self.my_model.id_lists_keys):
                id_list = self.my_model.id_lists[key]
                if id_list[0][self.my_model.idxs["type"]] != "blob":
                    # if i not in self.my_model.file_idxs:
                    print "setting index", i, "to hidden"
                    my_index = self.my_model.createIndex(i, 0, None)
                    index = self.model().mapFromSource(my_index)
                    self.setRowHidden(index.row(), QtCore.QModelIndex(), True)
        elif path_type == "tree":
            self.header().hideSection(self.my_model.idxs["type"])
            for i, key in enumerate(self.my_model.id_lists_keys):
                id_list = self.my_model.id_lists[key]
                if id_list[0][self.my_model.idxs["type"]] != "tree":
                    # if i not in self.my_model.dir_idxs:
                    print "setting index", i, "to hidden"
                    my_index = self.my_model.createIndex(i, 0, None)
                    index = self.model().mapFromSource(my_index)
                    self.setRowHidden(index.row(), QtCore.QModelIndex(), True)
        
    def set_selection(self):
        if self.current_id:
            my_index = self.my_model.find_row(self.current_id, 
                                              self.current_version)
            
            index = self.model().mapFromSource(my_index)
            if index.isValid():
                print 'checking internalPointer', my_index.internalPointer()
                if my_index.internalPointer():
                    my_expand_index = \
                        self.my_model.createIndex(
                        my_index.internalPointer().idx, 0, 0)
                    expand_index = self.model().mapFromSource(my_expand_index)
                    self.expand(expand_index)
                self.selectionModel().select(
                    index, QtGui.QItemSelectionModel.ClearAndSelect | \
                        QtGui.QItemSelectionModel.Rows)
                return True
        return False

    def set_id(self, id):
        self.current_id = id
        return self.set_selection()

    def set_version(self, version):
        self.current_version = version
        return self.set_selection()

    def get_id(self):
        sf_index = self.selectionModel().selectedRows()[0]
        index = self.model().mapToSource(sf_index)
        if index.internalPointer():
            paridx = index.internalPointer().idx
            id_list = \
                self.my_model.id_lists[self.my_model.id_lists_keys[paridx]]
            id = id_list[index.row()][self.my_model.idxs['id']]
        else:
            id_list = \
                self.my_model.id_lists[self.my_model.id_lists_keys[index.row()]]
            id = id_list[0][self.my_model.idxs['id']]
        return str(id)

    def get_version(self):
        sf_index = self.selectionModel().selectedRows()[0]
        index = self.model().mapToSource(sf_index)
        if index.internalPointer():
            paridx = index.internalPointer().idx
            id_list = \
                self.my_model.id_lists[self.my_model.id_lists_keys[paridx]]
            version = id_list[index.row()][self.my_model.idxs['version']]
            return str(version)        
        return None
        
    def get_info(self):
        return self.get_info_list()[0]

    def get_info_list(self):
        s_indexes = self.selectionModel().selectedRows()
        info_list = []
        for s_idx in s_indexes:
            index = self.model().mapToSource(s_idx)
            if index.internalPointer():
                paridx = index.internalPointer().idx
                id_list = \
                    self.my_model.id_lists[self.my_model.id_lists_keys[paridx]]
                info = id_list[index.row()]
                version = str(info[self.my_model.idxs['version']])
            else:
                id_list = self.my_model.id_lists[ \
                    self.my_model.id_lists_keys[index.row()]]
                info = id_list[0]
                version = None
            info_list.append((str(info[self.my_model.idxs['id']]),
                              version,
                              str(info[self.my_model.idxs['name']]),
                              str(info[self.my_model.idxs['tags']])))
        return info_list

class PersistentRefDialog(QtGui.QDialog):
    def __init__(self, param, parent=None):
        QtGui.QDialog.__init__(self, parent)
        # two tabs, one for starting from managed, one for local file
        # options are set accordingly
        # uuid assigned when options are set, either new or existing
        # don't compute uuid for local file that hasn't been run
        # need PersistentReference parameter
        # allow user to pass contents in to PersistentRef
        self.setWindowTitle("Configure Persistent Reference...")

        self.settings = {'ref_id': None,
                         'ref_version': None,
                         'local_path': None,
                         'versioning': False,
                         'local_read_priority': True,
                         'write_back': True}

        self.current_file = ""
        db_file = "/vistrails/managed/files.db"

        layout = QtGui.QVBoxLayout()
        managed_group = QtGui.QGroupBox("Persistent Data")
        managed_layout = QtGui.QVBoxLayout()
        search = QSearchBox(False, False)

        def keyPressEvent(obj, e):
            print "got to key press event", e.key()
            if e.key() in (QtCore.Qt.Key_Return,QtCore.Qt.Key_Enter):
                if obj.currentText():
                    obj.emit(QtCore.SIGNAL('executeSearch(QString)'),  
                             obj.searchEdit.currentText())
                else:
                    obj.emit(QtCore.SIGNAL('resetSearch()'))
            QtGui.QComboBox.keyPressEvent(obj, e)

        print 'keyPressEvent:', search.searchEdit.keyPressEvent
        search.searchEdit.keyPressEvent = keyPressEvent
        print 'keyPressEvent:', search.searchEdit.keyPressEvent
        self.connect(search, QtCore.SIGNAL('executeSearch(QString)'),
                     self.search_string)
        self.connect(search, QtCore.SIGNAL('resetSearch()'),
                     self.reset_search)
        managed_layout.addWidget(search)
        self.table = PersistentRefView(db_file, self)
        managed_layout.addWidget(self.table)
        managed_group.setLayout(managed_layout)
        layout.addWidget(managed_group)

        local_group = QtGui.QGroupBox("Local Data")
        local_layout = QtGui.QHBoxLayout()
        self.filename_edit = QtGui.QLineEdit()
        local_layout.addWidget(self.filename_edit)

        filename_button = QtGui.QToolButton()
        filename_button.setIcon(
            QtGui.QIcon(filename_button.style().standardPixmap(
                    QtGui.QStyle.SP_DirOpenIcon)))
        filename_button.setIconSize(QtCore.QSize(12,12))
        filename_button.connect(filename_button,
                                QtCore.SIGNAL('clicked()'),
                                self.choose_file)
        local_layout.addWidget(filename_button)
        local_group.setLayout(local_layout)
        layout.addWidget(local_group)

        pref_layout = QtGui.QHBoxLayout()
        version_group = QtGui.QGroupBox("Versioning")
        version_layout = QtGui.QVBoxLayout()
        version_off = QtGui.QRadioButton("Create New ID")
        version_layout.addWidget(version_off)
        version_on = QtGui.QRadioButton("Create New Version")
        version_layout.addWidget(version_on)
        version_group.setLayout(version_layout)
        pref_layout.addWidget(version_group)

        r_priority_group = QtGui.QGroupBox("Read Priority")
        r_priority_layout = QtGui.QVBoxLayout()
        r_priority_off = QtGui.QRadioButton("Local")
        r_priority_layout.addWidget(r_priority_off)
        r_priority_on = QtGui.QRadioButton("Persistent Store")
        r_priority_layout.addWidget(r_priority_on)
        r_priority_group.setLayout(r_priority_layout)
        pref_layout.addWidget(r_priority_group)
        
        w_priority_group = QtGui.QGroupBox("Write Priority")
        w_priority_layout = QtGui.QVBoxLayout()
        w_priority_off = QtGui.QRadioButton("Local")
        w_priority_layout.addWidget(w_priority_off)
        w_priority_on = QtGui.QRadioButton("Persistent Store")
        w_priority_layout.addWidget(w_priority_on)
        w_priority_group.setLayout(w_priority_layout)
        pref_layout.addWidget(w_priority_group)
        layout.addLayout(pref_layout)

        button_layout = QtGui.QHBoxLayout()
        button_layout.setDirection(QtGui.QBoxLayout.RightToLeft)
        button_layout.setAlignment(QtCore.Qt.AlignRight)
        ok_button = QtGui.QPushButton("OK")
        ok_button.setFixedWidth(100)
        self.connect(ok_button, QtCore.SIGNAL('clicked()'), self.close)
        button_layout.addWidget(ok_button)
        cancel_button = QtGui.QPushButton("Cancel")
        cancel_button.setFixedWidth(100)
        self.connect(cancel_button, QtCore.SIGNAL('clicked()'), self.cancel)
        button_layout.addWidget(cancel_button)
        layout.addLayout(button_layout)
        self.setLayout(layout)

    def close(self):
        self.done(QtGui.QDialog.Accepted)

    def cancel(self):
        self.done(QtGui.QDialog.Rejected)

    def choose_file(self):
        chosen_file = \
            QtGui.QFileDialog.getOpenFileName(self,
                                              'Use File...',
                                              self.current_file,
                                              'All files (*.*)')
        if chosen_file:
            self.current_file = chosen_file
            self.filename_edit.setText(self.current_file)
    
    def build_table(self):
        # read sql table from sqlite3 
        # display table
        pass
    
    def select_file(self):
        # present file dialog and allow user to choose file
        # compute file hash
        # if content exists in persistent store, ask to reuse 
        # (or set reuse checkbox)
        # set use_local checkbox to True
        # don't transfer until execution?
        # create uuid for file
        pass

    def select_remote(self):
        # present table of data with search box
        # each data has name, id, user, date, tags, etc.
        # if user selects, check use_local checkbox to False
        # store local is False, too?
        # don't transfer file until execution?
        # populate uuid
        pass


    # starting choice: local file or find existing managed file?
    # allow branch off of existing managed file -- ie no longer a
    # version of the initial file?
    
    # === options ===
    # overwrite managed on change
    # overwrite local on change
    # local is authoritative
    # keep local copy -- need to allow where to store

class PathChooserLayout(QtGui.QHBoxLayout):
    def __init__(self, is_dir=False, par_widget=None, parent=None):
        # QtGui.QWidget.__init__(self, parent)
        QtGui.QHBoxLayout.__init__(self, parent)
        self.par_widget = par_widget
        # layout = QtGui.QHBoxLayout()
        self.pathname_edit = QtGui.QLineEdit()
        self.addWidget(self.pathname_edit)

        pathname_button = QtGui.QToolButton()
        pathname_button.setIcon(
            QtGui.QIcon(pathname_button.style().standardPixmap(
                    QtGui.QStyle.SP_DirOpenIcon)))
        pathname_button.setIconSize(QtCore.QSize(12,12))
        pathname_button.connect(pathname_button,
                                QtCore.SIGNAL('clicked()'),
                                self.choose_path)
        self.addWidget(pathname_button)
        # layout.setContentsMargins(1,1,1,1)
        # self.setLayout(layout)
        self.is_dir = is_dir
        
    def choose_path(self):
        if self.is_dir:
            chosen_path = \
                QtGui.QFileDialog.getExistingDirectory(self.par_widget,
                                                       'Use Directory...',
                                                       self.pathname_edit.text())
        else:
            chosen_path = \
                QtGui.QFileDialog.getOpenFileName(self.par_widget,
                                                  'Use File...',
                                                  self.pathname_edit.text(),
                                                  'All files (*.*)')

        if chosen_path and chosen_path:
            self.pathname_edit.setText(chosen_path)
            self.emit(QtCore.SIGNAL('pathnameChanged()'))

    def get_path(self):
        return str(self.pathname_edit.text())

    def set_path(self, pathname):
        if pathname:
            self.pathname_edit.setText(pathname)
        else:
            self.pathname_edit.clear()

class PersistentRefViewSearch(QtGui.QGroupBox):
    def __init__(self, path_type=None, parent=None):
        QtGui.QGroupBox.__init__(self, parent)
        self.build_gui(path_type)

    def build_gui(self, path_type):
        layout = QtGui.QVBoxLayout()
        self.search_ref = QSearchBox(False, False)

        self.connect(self.search_ref, 
                     QtCore.SIGNAL('executeSearch(QString)'),
                     self.search_string)
        self.connect(self.search_ref, 
                     QtCore.SIGNAL('resetSearch()'),
                     self.reset_search)

        layout.addWidget(self.search_ref)
        self.ref_widget = PersistentRefView(path_type, self)
        layout.addWidget(self.ref_widget)
        layout.setMargin(0)
        self.setLayout(layout)

    def search_string(self, str):
        self.ref_widget.model().setFilterWildcard(str)

    def reset_search(self):
        self.ref_widget.model().setFilterWildcard('')
        self.ref_widget.model().invalidate()
    

class PersistentPathConfiguration(StandardModuleConfigurationWidget):
    def __init__(self, module, controller, parent=None, 
                 is_input=None, path_type=None):
        StandardModuleConfigurationWidget.__init__(self, module, controller, 
                                                   parent)

        # set title
        if module.has_annotation_with_key('__desc__'):
            label = module.get_annotation_by_key('__desc__').value.strip()
            title = '%s (%s) Module Configuration' % (label, module.name)
        else:
            title = '%s Module Configuration' % module.name
        self.setWindowTitle(title)

        self.build_gui(is_input, path_type)
        self.set_values()

    def sizeHint(self):
        return QtCore.QSize(800, 600)

    def build_gui(self, is_input, path_type):
        self.current_path = ""

        layout = QtGui.QVBoxLayout()
#         layout.setMargin(0)
#         layout.setSpacing(0)

        if not is_input:
            self.managed_change = \
                QtGui.QRadioButton("Always Create New Reference")
            layout.addWidget(self.managed_change)
            self.connect(self.managed_change, QtCore.SIGNAL("toggled(bool)"),
                         self.managed_toggle)

        else:
            self.managed_change = None

        self.managed_new = QtGui.QRadioButton("Create New Reference")
        self.connect(self.managed_new, QtCore.SIGNAL("toggled(bool)"),
                     self.new_toggle)
        layout.addWidget(self.managed_new)
        self.new_group = QtGui.QGroupBox()
        new_layout = QtGui.QGridLayout()
        self.new_file = None
        if is_input:
            new_layout.addWidget(QtGui.QLabel("Path:"), 0, 0)
            self.new_file = self.get_chooser_layout()
            if hasattr(self.new_file, 'pathname_edit'):
                self.connect(self.new_file.pathname_edit,
                             QtCore.SIGNAL("textChanged(QString)"),
                             self.stateChange)
            new_layout.addLayout(self.new_file, 0, 1)
            self.connect(self.new_file, QtCore.SIGNAL("pathnameChanged()"),
                         self.new_file_changed)
        new_layout.addWidget(QtGui.QLabel("Name:"), 1, 0)
        self.name_edit = QtGui.QLineEdit()
        self.connect(self.name_edit, QtCore.SIGNAL("textChanged(QString)"),
                     self.stateChange)
        new_layout.addWidget(self.name_edit, 1, 1)
        new_layout.addWidget(QtGui.QLabel("Tags:"), 2, 0)
        self.tags_edit = QtGui.QLineEdit()
        self.connect(self.tags_edit, QtCore.SIGNAL("textChanged(QString)"),
                     self.stateChange)
        new_layout.addWidget(self.tags_edit, 2, 1)
        self.new_group.setLayout(new_layout)
        layout.addWidget(self.new_group)

        self.managed_existing = QtGui.QRadioButton("Use Existing Reference")
        self.connect(self.managed_existing, QtCore.SIGNAL("toggled(bool)"),
                     self.existing_toggle)
        layout.addWidget(self.managed_existing)

        # self.existing_group = QtGui.QGroupBox()
        # existing_layout = QtGui.QVBoxLayout()
        # self.search_ref = QSearchBox(False, False)

        # self.connect(self.search_ref, 
        #              QtCore.SIGNAL('executeSearch(QString)'),
        #              self.search_string)
        # self.connect(self.search_ref, 
        #              QtCore.SIGNAL('resetSearch()'),
        #              self.reset_search)

        # existing_layout.addWidget(self.search_ref)
        # self.ref_widget = PersistentRefView(path_type, self)
        # existing_layout.addWidget(self.ref_widget)
        # self.existing_group.setLayout(existing_layout)
        self.existing_group = PersistentRefViewSearch(path_type)
        self.ref_widget = self.existing_group.ref_widget
        self.connect(self.ref_widget,
                     QtCore.SIGNAL("clicked(QModelIndex)"),
                     self.ref_changed)
        layout.addWidget(self.existing_group)

        self.keep_local = QtGui.QCheckBox("Keep Local Version")
        layout.addWidget(self.keep_local)
        self.connect(self.keep_local, QtCore.SIGNAL("toggled(bool)"),
                     self.local_toggle)
        self.local_group = QtGui.QGroupBox()
        local_layout = QtGui.QGridLayout()
        self.local_path = self.get_chooser_layout()
        if hasattr(self.local_path, 'pathname_edit'):
            self.connect(self.local_path.pathname_edit,
                         QtCore.SIGNAL("textChanged(QString)"),
                         self.stateChange)
        local_layout.addLayout(self.local_path,0,0,1,2)

        self.r_priority_local = QtGui.QCheckBox("Read From Local Path")
        local_layout.addWidget(self.r_priority_local,1,0)
        self.write_managed_checkbox = QtGui.QCheckBox("Write To Local Path")
        self.connect(self.write_managed_checkbox, QtCore.SIGNAL("toggled(bool)"),
                     self.stateChange)
        local_layout.addWidget(self.write_managed_checkbox,1,1)
        self.local_group.setLayout(local_layout)
        layout.addWidget(self.local_group)

        if is_input:
            self.r_priority_local.setEnabled(True)
            self.write_managed_checkbox.setEnabled(False)
        else:
            self.r_priority_local.setEnabled(False)
            self.write_managed_checkbox.setEnabled(True)

        button_layout = QtGui.QHBoxLayout()
        button_layout.setDirection(QtGui.QBoxLayout.RightToLeft)
        button_layout.setAlignment(QtCore.Qt.AlignRight)
        self.saveButton = QtGui.QPushButton("Save")
        self.saveButton.setFixedWidth(100)
        self.connect(self.saveButton, QtCore.SIGNAL('clicked(bool)'),
                     self.saveTriggered)
        button_layout.addWidget(self.saveButton)
        self.resetButton = QtGui.QPushButton("Reset")
        self.resetButton.setFixedWidth(100)
        self.connect(self.resetButton, QtCore.SIGNAL('clicked()'),
                     self.resetTriggered)
        button_layout.addWidget(self.resetButton)
        layout.addLayout(button_layout)
        self.setLayout(layout)

    def saveTriggered(self, checked = False):
        self.get_values()
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)
        self.state_changed = False
        self.emit(QtCore.SIGNAL('doneConfigure'), self.module.id)
        
    def closeEvent(self, event):
        self.askToSaveChanges()
        event.accept()
                
    def resetTriggered(self):
        self.set_values()
        self.setUpdatesEnabled(True)
        self.state_changed = False
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)
        
    def stateChange(self, checked = False, old = None):
        if not self.state_changed:
            self.state_changed = True
            self.saveButton.setEnabled(True)
            self.resetButton.setEnabled(True)
    
    def managed_toggle(self, checked):
        self.stateChange()
        self.new_group.setEnabled(not checked)
        self.existing_group.setEnabled(not checked)

    def new_toggle(self, checked):
        self.stateChange()
        self.new_group.setEnabled(checked)
        self.existing_group.setEnabled(not checked)
        if not checked and self.keep_local.isChecked():
            self.keep_local.setChecked(False)

    def existing_toggle(self, checked):
        self.stateChange()
        self.existing_group.setEnabled(checked)
        self.new_group.setEnabled(not checked)

    def local_toggle(self, checked):
        self.stateChange()
        self.local_group.setEnabled(checked)

    def new_file_changed(self):
        self.stateChange()
        new_file = str(self.new_file.get_path())
        if new_file:
            base_name = os.path.basename(new_file)
        else:
            base_name = ''
        self.name_edit.setText(base_name)
        self.keep_local.setChecked(True)
        self.local_path.set_path(new_file)
        self.r_priority_local.setChecked(True)
        self.write_managed_checkbox.setChecked(False)

    def ref_changed(self, index):
        self.stateChange()
        if self.keep_local.isChecked():
            self.keep_local.setChecked(False)

    def set_values(self):
        from vistrails.core.modules.module_registry import get_module_registry
        reg = get_module_registry()
        PersistentRef = \
            reg.get_descriptor_by_name(persistence_pkg, 'PersistentRef').module

        def func_to_bool(function):
            try:
                value = function.parameters[0].strValue
            except IndexError:
                return False
            if value and value == 'True':
                return True

        ref_exists = False
        self.existing_ref = None
        local_path = None
        local_read = None
        local_write = None
        for function in self.module.functions:
            if function.name == 'ref':
                self.existing_ref = PersistentRef.translate_to_python(
                    function.parameters[0].strValue)
                self.ref_widget.set_id(self.existing_ref.id)
                ref_exists = \
                    self.ref_widget.set_version(self.existing_ref.version)
                self.existing_ref._exists = ref_exists
                print 'ref_exists:', ref_exists, self.existing_ref.id, \
                    self.existing_ref.version
            elif function.name == 'value':
                if self.new_file:
                    self.new_file.set_path(function.parameters[0].strValue)
            elif function.name == 'localPath':
                local_path = Path.translate_to_python(
                    function.parameters[0].strValue).name
            elif function.name == 'readLocal':
                local_read = func_to_bool(function)
            elif function.name == 'writeLocal':
                local_write = func_to_bool(function)

        if ref_exists:
            self.managed_existing.setChecked(True)
            self.existing_toggle(True)
        elif self.managed_change and (not self.existing_ref or \
                                          not self.existing_ref.id):
            self.managed_change.setChecked(True)
            self.managed_toggle(True)
        else:
            self.managed_new.setChecked(True)
            self.new_toggle(True)
            if self.existing_ref:
                self.name_edit.setText(self.existing_ref.name)
                self.tags_edit.setText(self.existing_ref.tags)
        
        if self.existing_ref:
            if self.existing_ref.local_path:
                self.keep_local.setChecked(True)
                self.local_toggle(True)
            else:
                self.keep_local.setChecked(False)
                self.local_toggle(False)
            self.local_path.set_path(self.existing_ref.local_path)
            self.r_priority_local.setChecked(self.existing_ref.local_read)
            self.write_managed_checkbox.setChecked(
                self.existing_ref.local_writeback)
        else:
            self.keep_local.setChecked(False)
            self.local_toggle(False)

        if local_path is not None:
            if local_path:
                self.keep_local.setChecked(True)
                self.local_toggle(True)
            else:
                self.keep_local.setChecked(False)
                self.local_toggle(False)
            self.local_path.set_path(local_path)
        if local_read is not None:
            self.r_priority_local.setChecked(local_read)
        if local_write is not None:
            self.write_managed_checkbox.setChecked(local_write)
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)
        self.state_changed = False

    def get_values(self):
        from vistrails.core.modules.module_registry import get_module_registry
        reg = get_module_registry()
        PersistentRef = \
            reg.get_descriptor_by_name(persistence_pkg, 'PersistentRef').module

        functions = []
        if self.new_file and self.new_file.get_path() and \
                self.managed_new.isChecked():
        # if self.new_file and self.new_file.get_path():
            functions.append(('value', [self.new_file.get_path()]))
        else:
            functions.append(('value', None))
            pass
        ref = PersistentRef()
        if self.managed_new.isChecked():
            if self.existing_ref and not self.existing_ref._exists:
                ref.id = self.existing_ref.id
                ref.version = self.existing_ref.version
            else:
                ref.id = str(uuid.uuid1())
                ref.version = None
            # endif
            ref.name = str(self.name_edit.text())
            ref.tags = str(self.tags_edit.text())
        elif self.managed_existing.isChecked():
            (ref.id, ref.version, ref.name, ref.tags) = \
                self.ref_widget.get_info()
        if self.keep_local.isChecked():
            functions.append(('localPath', [self.local_path.get_path()]))
            functions.append(('readLocal', 
                              [str(self.r_priority_local.isChecked())]))
            functions.append(('writeLocal',
                              [str(self.write_managed_checkbox.isChecked())]))
#             ref.local_path = self.local_path.get_path()
#             ref.local_read = self.r_priority_local.isChecked()
#             ref.local_writeback = self.write_managed_checkbox.isChecked()
        else:
            ref.local_path = None
            # functions.append(('localPath', None))
            functions.append(('readLocal', None))
            functions.append(('writeLocal', None))
            pass
        functions.append(('ref', [PersistentRef.translate_to_string(ref)]))
        self.controller.update_functions(self.module, functions)

class PersistentInputPathConfiguration(PersistentPathConfiguration):
    def __init__(self, module, controller, parent=None, path_type=None):
        PersistentPathConfiguration.__init__(self, module, controller, parent,
                                          True, path_type)
        
class PersistentOutputPathConfiguration(PersistentPathConfiguration):
    def __init__(self, module, controller, parent=None, path_type=None):
        PersistentPathConfiguration.__init__(self, module, controller, parent,
                                          False, path_type)

class PersistentRefInlineWidget(QtGui.QWidget, ConstantWidgetMixin):
    contentsChanged = QtCore.pyqtSignal(tuple)
    def __init__(self, param, parent=None):
        self.param = param
        self.strValue = param.strValue
        contentsType = param.type
        QtGui.QWidget.__init__(self, parent)
        ConstantWidgetMixin.__init__(self, param.strValue)
        layout = QtGui.QHBoxLayout()
        # FIXME Use a greyed QLineEdit?
        # layout.addWidget(QtGui.QLabel("File Info:"))
        button = QtGui.QPushButton("Configure")
        button.setMaximumWidth(100)
        self.connect(button, QtCore.SIGNAL('clicked()'), self.run_dialog)
        layout.addWidget(button)
        layout.setMargin(5)
        layout.setSpacing(5)
        self.setLayout(layout)
        
    def run_dialog(self):
        dialog = PersistentRefDialog(self.param)
        if dialog.exec_() == QtGui.QDialog.Accepted:
            self.setContents("test")
            #use same translate call?, False)

    def contents(self):
        return self.strValue

    def setContents(self, strValue, silent=True):
        self.strValue = strValue
        if not silent:
            self.update_parent()

class PersistentInputFileConfiguration(PersistentInputPathConfiguration):
    def __init__(self, module, controller, parent=None):
        PersistentInputPathConfiguration.__init__(self, module, controller, 
                                               parent, "blob")
    def get_chooser_layout(self):
        return PathChooserLayout(False, self)
    
class PersistentInputDirConfiguration(PersistentInputPathConfiguration):
    def __init__(self, module, controller, parent=None):
        PersistentInputPathConfiguration.__init__(self, module, controller, 
                                               parent, "tree")

    def get_chooser_layout(self):
        return PathChooserLayout(True, self)

class PersistentOutputFileConfiguration(PersistentOutputPathConfiguration):
    def __init__(self, module, controller, parent=None):
        PersistentOutputPathConfiguration.__init__(self, module, controller, 
                                               parent, "blob")

    def get_chooser_layout(self):
        return PathChooserLayout(False, self)

class PersistentOutputDirConfiguration(PersistentOutputPathConfiguration):
    def __init__(self, module, controller, parent=None):
        PersistentOutputPathConfiguration.__init__(self, module, controller, 
                                               parent, "tree")

    def get_chooser_layout(self):
        return PathChooserLayout(True, self)
    
class PersistentConfiguration(QtGui.QDialog):
    def __init__(self, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.setModal(False)
        self.build_gui()
        self.db_access = DatabaseAccessSingleton()

    def build_gui(self):
        layout = QtGui.QVBoxLayout()
        self.ref_search = PersistentRefViewSearch(None)
        self.ref_search.ref_widget.setSelectionMode(
            QtGui.QAbstractItemView.ExtendedSelection)
        layout.addWidget(self.ref_search)
        
        button_layout = QtGui.QHBoxLayout()
        button_layout.setAlignment(QtCore.Qt.AlignRight)
        write_button = QtGui.QPushButton("Write...")
        write_button.setAutoDefault(False)
        self.connect(write_button, QtCore.SIGNAL("clicked()"), self.write)
        button_layout.addWidget(write_button)
        delete_button = QtGui.QPushButton("Delete...")
        delete_button.setAutoDefault(False)
        self.connect(delete_button, QtCore.SIGNAL("clicked()"), self.delete)
        button_layout.addWidget(delete_button)
        layout.addLayout(button_layout)
        self.setLayout(layout)
        
    def sizeHint(self):
        return QtCore.QSize(800,320)

    def write(self):
        info_list = self.ref_search.ref_widget.get_info_list()
        if len(info_list) < 1:
            return
        elif len(info_list) == 1:
            # save single file/dir
            info = info_list[0]
            name = info[2]
            chosen_path = QtGui.QFileDialog.getSaveFileName(
                    self,
                    'Save...',
                    name)
            if not chosen_path:
                return

            # FIXME really should move this calls to a higher level so
            # we don't need to instantiate a module
            if info[1] is None:
                version = "HEAD"
            else:
                version = info[1]
            repo.get_current_repo().get_path(info[0], version, None, 
                                             chosen_path)
        else:
            # have multiple files/dirs
            chosen_path = QtGui.QFileDialog.getExistingDirectory(
                    self,
                    'Save All to Directory...')
            has_overwrite = False
            # if untitled (no name, use the uuid)
            for info in info_list:
                if info[2]:
                    name = info[2]
                else:
                    name = info[0]
                full_path = os.path.join(chosen_path, name)
                if os.path.exists(full_path):
                    has_overwrite = True
            if has_overwrite:
                question_str = "One or more of the paths already exist.  " + \
                    "Overwrite?"
                ret_val = \
                    QtGui.QMessageBox.question(self, "Overwrite", \
                                                   question_str, \
                                                   QtGui.QMessageBox.Cancel | \
                                                   QtGui.QMessageBox.No | \
                                                   QtGui.QMessageBox.Yes)
                if ret_val != QtGui.QMessageBox.Yes:
                    return

            for info in info_list:
                if info[1] is None:
                    version = "HEAD"
                else:
                    version = info[1]
                if info[2]:
                    name = info[2]
                else:
                    name = info[0]
                full_path = os.path.join(chosen_path, name)
                repo.get_current_repo().git_get_path(info[0], version, None, 
                                                     full_path)
            
    def delete(self):
        QtGui.QMessageBox.critical(self, "Delete",
                                   "This feature is not functional in the "
                                   "current version of VisTrails and has been "
                                   "disabled for this release.")
        return

        from init import PersistentPath
        info_list = self.ref_search.ref_widget.get_info_list()
        if len(info_list) < 1:
            return

        delete_str = "This will permanently delete the selected data " + \
            "from the peristent store.  This cannot be undone.  Proceed?"
        question_f = QtGui.QMessageBox.question
        ret_val = question_f(self, "Delete", delete_str, \
                                 QtGui.QMessageBox.Cancel | \
                                 QtGui.QMessageBox.Ok)
        if ret_val != QtGui.QMessageBox.Ok:
            return
            
        git_util = PersistentPath()
        db_access = DatabaseAccessSingleton()
        # FIXME keep entry in database with flag for deleted?
        # NEED TO update the model...
        for info in info_list:
            delete_where = {'id': info[0]}
            if info[1] is None:
                git_util.git_remove_path(info[0])
                db_access.delete_from_database(delete_where)
            else:
                # FIXME implement delete for versions...
                delete_where['version'] = info[1]
                print "NOT IMPLEMENTED FOR VERSIONS!!"
                
                
        



#!/usr/bin/env python

# This example shows how to use a transparent texture map to perform
# thresholding. The key is the vtkThresholdTextureCoords filter which
# creates texture coordinates based on a threshold value. These
# texture coordinates are used in conjuntion with a texture map with
# varying opacity and intensity to create an inside, transition, and
# outside region.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Begin by reading some structure grid data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/bluntfinxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/bluntfinq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# Now extract surfaces from the grid corresponding to boundary
# geometry.  First the wall.
wall = vtk.vtkStructuredGridGeometryFilter()
wall.SetInputConnection(pl3d.GetOutputPort())
wall.SetExtent(0, 100, 0, 0, 0, 100)
wallMap = vtk.vtkPolyDataMapper()
wallMap.SetInputConnection(wall.GetOutputPort())
wallMap.ScalarVisibilityOff()
wallActor = vtk.vtkActor()
wallActor.SetMapper(wallMap)
wallActor.GetProperty().SetColor(0.8, 0.8, 0.8)

# Now the fin.
fin = vtk.vtkStructuredGridGeometryFilter()
fin.SetInputConnection(pl3d.GetOutputPort())
fin.SetExtent(0, 100, 0, 100, 0, 0)
finMap = vtk.vtkPolyDataMapper()
finMap.SetInputConnection(fin.GetOutputPort())
finMap.ScalarVisibilityOff()
finActor = vtk.vtkActor()
finActor.SetMapper(finMap)
finActor.GetProperty().SetColor(0.8, 0.8, 0.8)

# Extract planes to threshold. Start by reading the specially designed
# texture map that has three regions: an inside, boundary, and outside
# region. The opacity and intensity of this texture map are varied.
tmap = vtk.vtkStructuredPointsReader()
tmap.SetFileName(VTK_DATA_ROOT + "/Data/texThres2.vtk")
texture = vtk.vtkTexture()
texture.SetInputConnection(tmap.GetOutputPort())
texture.InterpolateOff()
texture.RepeatOff()

# Here are the three planes which will be texture thresholded.
plane1 = vtk.vtkStructuredGridGeometryFilter()
plane1.SetInputConnection(pl3d.GetOutputPort())
plane1.SetExtent(10, 10, 0, 100, 0, 100)
thresh1 = vtk.vtkThresholdTextureCoords()
thresh1.SetInputConnection(plane1.GetOutputPort())
thresh1.ThresholdByUpper(1.5)
plane1Map = vtk.vtkDataSetMapper()
plane1Map.SetInputConnection(thresh1.GetOutputPort())
plane1Map.SetScalarRange(pl3d.GetOutput().GetScalarRange())
plane1Actor = vtk.vtkActor()
plane1Actor.SetMapper(plane1Map)
plane1Actor.SetTexture(texture)
plane1Actor.GetProperty().SetOpacity(0.999)

plane2 = vtk.vtkStructuredGridGeometryFilter()
plane2.SetInputConnection(pl3d.GetOutputPort())
plane2.SetExtent(30, 30, 0, 100, 0, 100)
thresh2 = vtk.vtkThresholdTextureCoords()
thresh2.SetInputConnection(plane2.GetOutputPort())
thresh2.ThresholdByUpper(1.5)
plane2Map = vtk.vtkDataSetMapper()
plane2Map.SetInputConnection(thresh2.GetOutputPort())
plane2Map.SetScalarRange(pl3d.GetOutput().GetScalarRange())
plane2Actor = vtk.vtkActor()
plane2Actor.SetMapper(plane2Map)
plane2Actor.SetTexture(texture)
plane2Actor.GetProperty().SetOpacity(0.999)

plane3 = vtk.vtkStructuredGridGeometryFilter()
plane3.SetInputConnection(pl3d.GetOutputPort())
plane3.SetExtent(35, 35, 0, 100, 0, 100)
thresh3 = vtk.vtkThresholdTextureCoords()
thresh3.SetInputConnection(plane3.GetOutputPort())
thresh3.ThresholdByUpper(1.5)
plane3Map = vtk.vtkDataSetMapper()
plane3Map.SetInputConnection(thresh3.GetOutputPort())
plane3Map.SetScalarRange(pl3d.GetOutput().GetScalarRange())
plane3Actor = vtk.vtkActor()
plane3Actor.SetMapper(plane3Map)
plane3Actor.SetTexture(texture)
plane3Actor.GetProperty().SetOpacity(0.999)

# For context create an outline around the data.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)
outlineProp = outlineActor.GetProperty()
outlineProp.SetColor(0, 0, 0)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(wallActor)
ren.AddActor(finActor)
ren.AddActor(plane1Actor)
ren.AddActor(plane2Actor)
ren.AddActor(plane3Actor)
ren.SetBackground(1, 1, 1)
renWin.SetSize(500, 500)

# Set up a nice view.
cam1 = vtk.vtkCamera()
cam1.SetClippingRange(1.51176, 75.5879)
cam1.SetFocalPoint(2.33749, 2.96739, 3.61023)
cam1.SetPosition(10.8787, 5.27346, 15.8687)
cam1.SetViewAngle(30)
cam1.SetViewUp(-0.0610856, 0.987798, -0.143262)
ren.SetActiveCamera(cam1)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

import sys
import os
import re

def run(in_dir, out_dir):
    def collectFilenames(dir):
        # create the regular expression matcher machines
        fileNameParser = re.compile('.*\.py')

        result = []
        for file in os.listdir(dir):
            # print childDirOrFile
            childDirOrFile = os.path.join(dir,file)
            if os.path.isfile(childDirOrFile):
                # file does match?
                if fileNameParser.match(childDirOrFile):
                    result.append(childDirOrFile)
            elif os.path.isdir(childDirOrFile):
                result += collectFilenames(childDirOrFile)
        return result

    
    all_files = collectFilenames(in_dir)

    in_base = os.path.basename(in_dir)
    out_base = os.path.basename(out_dir)
    for fname in all_files:
        path_before = os.path.dirname(fname)
        path_end = os.path.basename(fname)
        while os.path.basename(path_before) != in_base:
            path_end = os.path.join(os.path.basename(path_before), path_end)
            path_before = os.path.dirname(path_before)
        other_fname = os.path.join(out_dir, path_end)

        cmd_line = "diff %s %s" % (fname, other_fname)
        print cmd_line
        os.system(cmd_line)

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print 'Usage: python %s <in_directory> <out_directory>' % sys.argv[0]
        sys.exit(-1)
    run(*sys.argv[1:])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import os
import shutil
import sys

from compute_hash import compute_hash
from identifiers import identifier as persistence_pkg, \
    old_identifiers as persistence_old_ids
persistence_pkg_ids = set(persistence_old_ids)
persistence_pkg_ids.add(persistence_pkg)

vistrails_src = None
if not vistrails_src:
    vistrails_src = os.path.dirname(
        os.path.dirname(os.path.dirname(sys.path[0])))
if vistrails_src not in sys.path:
    sys.path.append(vistrails_src)
    
import vistrails.db.services.io    

def find_workflows(path_name, vistrail_dir):
    file_hash = compute_hash(path_name)
    vt_files = []
    dir_stack = [vistrail_dir]
    while dir_stack:
        dir = dir_stack.pop()
        for base in os.listdir(dir):
            name = os.path.join(dir, base)
            if os.path.isdir(name):
                dir_stack.append(name)
            elif name.endswith('.vt'):
                vt_files.append(name)

    vt_finds = {}
    for filename in vt_files:
        save_bundle, save_dir = \
            vistrails.db.services.io.open_vistrail_bundle_from_zip_xml(filename)
        vistrail = save_bundle.vistrail
        log_fname = vistrail.db_log_filename
        log = vistrails.db.services.io.open_log_from_xml(log_fname, True)
        
        persistent_module_ids = set()
        for action in vistrail.db_actions:
            for op in action.db_operations:
                if op.db_what == 'module' and (op.vtType == 'add' or 
                                               op.vtType == 'change'):
                    module = op.db_data
                    if module.db_package in persistence_pkg_ids:
                        persistent_module_ids.add(module.db_id)

        execs = {}
        tags = {}
        for workflow_exec in log.db_workflow_execs:
            cur_version = workflow_exec.db_parent_version
            if cur_version in vistrail.db_tags_id_index:
                tags[cur_version] = \
                    vistrail.db_tags_id_index[cur_version].db_name

            for module_exec in workflow_exec.db_item_execs:
                if module_exec.db_module_id in persistent_module_ids:
                    found = False
                    for annotation in module_exec.db_annotations:
                        if annotation.db_key == 'sha_hash':
                            if annotation.db_value == file_hash:
                                # found the file/dir
                                # do something
                                found = True
                        if annotation.db_key == 'signature':
                            val = annotation.db_value.upper()
                            cache_fname = os.path.join(val[:2], val[2:])
                    if found:
                        if cur_version not in execs:
                            execs[cur_version] = []
                        execs[cur_version].append((workflow_exec.db_ts_start,
                                                   cache_fname))

        if len(execs) > 0:
            vt_finds[filename] = (execs, tags)

        shutil.rmtree(save_dir)

    return vt_finds

if __name__ == '__main__':
    path_name = sys.argv[1]
    vistrail_dir = sys.argv[2]
    vt_finds = find_workflows(path_name, vistrail_dir)

    for fname, (execs, tags) in vt_finds.iteritems():
        print fname + ':'
        for version, exec_list in execs.iteritems():
            for (exec_time, cache_fname) in exec_list:
                if version in tags:
                    print ' ', str(version) + ' (' + tags[version] + '):', \
                        exec_time
                    print '   ', cache_fname
                else:
                    print ' ', str(version) + ':', exec_time
                    print '   ', cache_fname



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

""" Convert VTK classes into functions using spec in vtk.xml"""

from __future__ import division

import locale
import os
import tempfile
import types

from  platform import system

import vtk

from . import fix_classes
from .wrapper import VTKInstanceWrapper
from .specs import SpecList, ClassSpec

################################################################################

# filter some deprecation warnings coming from the fact that vtk calls
# range() with float parameters

#### METHOD PATCHING CODE ####

def patch_methods(base_module, cls):
    """class_dict(base_module, cls: vtkClass) -> dict
    Returns the class dictionary for the module represented by base_module and
    with base class base_module"""

    instance_dict = {}
    def update_dict(name, callable_):
        if instance_dict.has_key(name):
            instance_dict[name] = callable_(types.MethodType(instance_dict[name], base_module))
        elif hasattr(base_module, name):
            instance_dict[name] = callable_(getattr(base_module, name))
        else:
            instance_dict[name] = callable_(None)

    def compute_UpdateAlgorithm(oldUpdate):
        def call_UpdateAlgorithm(self):
            if self._callback is None:
                oldUpdate()
                return
            is_aborted = [False]
            cbId = None
            def ProgressEvent(obj, event):
                try:
                    self._callback(obj.GetProgress())
                except Exception, e:
                    if e.__name__ == 'AbortExecution':
                        obj.SetAbortExecute(True)
                        self.RemoveObserver(cbId)
                        is_aborted[0] = True
                    else:
                        raise
            cbId = self.AddObserver('ProgressEvent', ProgressEvent)
            oldUpdate()
            if not is_aborted[0]:
                self.RemoveObserver(cbId)
        return call_UpdateAlgorithm
    if issubclass(cls, vtk.vtkAlgorithm):
        update_dict('Update', compute_UpdateAlgorithm)

    def guarded_SimpleScalarTree_wrap_compute(old_compute):
        # This builds the scalar tree
        def compute(self):
            old_compute(self)
            self.vtkInstance.BuildTree()
        return compute
    if issubclass(cls, vtk.vtkScalarTree):
        update_dict('Update', guarded_SimpleScalarTree_wrap_compute)

    def guarded_Writer_wrap_compute(self):
        # The behavior for vtkWriter subclasses is to call Write()
        # If the user sets a name, we will create a file with that name
        # If not, we will create a temporary file using _tempfile
        fn = self.vtkInstance.GetFileName()
        if not fn:
            fn = self._tempfile(suffix='.vtk')
            self.vtkInstance.SetFileName(fn)
        self.vtkInstance.Write()
        return fn
    if issubclass(cls, vtk.vtkWriter):
        instance_dict['file'] = guarded_Writer_wrap_compute

    def guarded_SetFileName(old_compute):
        # This builds the scalar tree
        def check_SetFileName(self):
            # This checks for the presence of file in VTK readers
            # Skips the check if it's a vtkImageReader or vtkPLOT3DReader, because
            # it has other ways of specifying files, like SetFilePrefix for
            # multiple files
            skip = [vtk.vtkBYUReader,
                    vtk.vtkImageReader,
                    vtk.vtkDICOMImageReader,
                    vtk.vtkTIFFReader]
            # vtkPLOT3DReader does not exist from version 6.0.0
            v = vtk.vtkVersion()
            version = [v.GetVTKMajorVersion(),
                       v.GetVTKMinorVersion(),
                       v.GetVTKBuildVersion()]
            if version < [6, 0, 0]:
                skip.append(vtk.vtkPLOT3DReader)
            if not any(issubclass(cls, x) for x in skip):
                filename = self.vtkInstance.GetFileName()
                if not os.path.isfile(filename):
                    raise Exception('File does not exist')
            old_compute()
        return check_SetFileName
    if hasattr(cls, 'SetFileName') and \
       cls.__name__.endswith('Reader') and \
       not cls.__name__.endswith('TiffReader'):
        update_dict('Update', guarded_SetFileName)

    def call_SetRenderWindow(self, vtkRenderer):
        window = vtk.vtkRenderWindow()
        w = 512
        h = 512
        window.OffScreenRenderingOn()
        window.SetSize(w, h)

        widget = None
        if system() == 'Darwin':
            from PyQt4 import QtCore, QtGui
            widget = QtGui.QWidget(None, QtCore.Qt.FramelessWindowHint)
            widget.resize(w, h)
            widget.show()
            window.SetWindowInfo(str(int(widget.winId())))

        window.AddRenderer(vtkRenderer.vtkInstance)
        window.Render()
        self.vtkInstance.SetRenderWindow(window)
    if hasattr(cls, 'SetRenderWindow'):
        instance_dict['vtkRenderer'] = call_SetRenderWindow

    def call_TransferFunction(self, tf):
        tf.set_on_vtk_volume_property(self.vtkInstance)
    if issubclass(cls, vtk.vtkVolumeProperty):
        instance_dict['SetTransferFunction'] = call_TransferFunction

    def call_PointData(self, pd):
        self.vtkInstance.GetPointData().ShallowCopy(pd.vtkInstance)
    def call_CellData(self, cd):
        self.vtkInstance.GetCellData().ShallowCopy(cd.vtkInstance)
    if issubclass(cls, vtk.vtkDataSet):
        instance_dict['PointData'] = call_PointData
        instance_dict['CellData'] = call_CellData

    def call_PointIds(self, point_ids):
        self.vtkInstance.GetPointIds().SetNumberOfIds(point_ids.GetNumberOfIds())
        for i in xrange(point_ids.GetNumberOfIds()):
            self.vtkInstance.GetPointIds().SetId(i, point_ids.vtkInstance.GetId(i))
    if issubclass(cls, vtk.vtkCell):
        instance_dict['PointIds'] = call_PointIds

    def call_CopyImportVoidPointer(self, pointer):
        self.CopyImportVoidPointer(pointer, len(pointer))
        return call_CopyImportVoidPointer
    if issubclass(cls, vtk.vtkImageImport):
        instance_dict['CopyImportString'] = call_CopyImportVoidPointer

    def call_GetFirstBlock(self):
        return VTKInstanceWrapper(self.vtkInstance.GetOutput().GetBlock(0))
    if issubclass(cls, vtk.vtkMultiBlockPLOT3DReader):
        instance_dict['FirstBlock'] = call_GetFirstBlock

    for name, method in instance_dict.iteritems():
        setattr(base_module, name, types.MethodType(method, base_module))

#### END METHOD PATCHING CODE ####


class vtkObjectInfo(object):
    """ Each instance represents a VTK class

    """
    def __init__(self, spec, parent=None):
        self.parent = parent
        self.spec = spec
        self.vtkClass = getattr(vtk, spec.code_ref)
        # use fixed classes
        if hasattr(fix_classes, self.vtkClass.__name__ + '_fixed'):
            self.vtkClass = getattr(fix_classes, self.vtkClass.__name__ + '_fixed')

class VTKInstancePatcher(object):
    """ This is used in place of the vtk instance because it may not be
        safe to set attributes directly on the vtk object.

    """
    def __init__(self, info_obj):
        # fixes reading data files on non-C locales
        self._previous_locale = locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_ALL, 'C')

        self._info_obj = info_obj
        self._callback = None
        self._tempfile = tempfile.mkstemp
        self.vtkInstance = info_obj.vtkClass()
        patch_methods(self, info_obj.vtkClass)

    def _cleanup(self):
        locale.setlocale(locale.LC_ALL, self._previous_locale)

    def __getattr__(self, name):
        v = vtk.vtkVersion()
        version = [v.GetVTKMajorVersion(),
                   v.GetVTKMinorVersion(),
                   v.GetVTKBuildVersion()]
        if version < [6, 0, 0]:
            # Translate vtk6 method names to vtk5 method names
            to_vtk5_names = {'AddInputData':  'AddInput',
                          'AddDataSetInput':  'AddInput',
                             'SetInputData':  'SetInput',
                            'AddSourceData': 'AddSource',
                            'SetSourceData': 'SetSource'}
            for new, old in to_vtk5_names.iteritems():
                if name.startswith(new):
                    # Keep suffix
                    name = old + name[len(new):]
                    break
        # redirect calls to vtkInstance
        def call_wrapper(*args):
            args = list(args)
            for i in xrange(len(args)):
                if hasattr(args[i], 'vtkInstance'):
                    # Unwrap VTK objects
                    args[i] = args[i].vtkInstance
            args = tuple(args)

            #print "CALLING", name, [a.__class__.__name__ for a in args]
            result = getattr(self.vtkInstance, name)(*args)
            #print "GOT", result.__class__.__name__

            if isinstance(result, vtk.vtkObjectBase):
                # Wrap VTK objects
                result = VTKInstanceWrapper(result)
            return result
        return call_wrapper

    def _set_callback(self, callback):
        self._callback = callback

    def _set_tempfile(self, tempfile):
        self._tempfile = tempfile
        
    def Update(self):
        if hasattr(self.vtkInstance, 'Update'):
            self.vtkInstance.Update()

# keep track of created modules for use as subclasses
infoObjs = {}


def gen_instance_factory(spec):
    """Create an instance factory from a vtk class specification

    """
    infoObj = vtkObjectInfo(spec, infoObjs.get(spec.superklass, None))
    infoObjs[spec.module_name] = infoObj
    def instanceFactory():
        return VTKInstancePatcher(infoObj)
    return instanceFactory


specs = None


def initialize(spec_name=None):
    """ Generate class wrappers and add them to current module namespace
        Also adds spec so it can be referenced by module wrapper

    """
    global specs
    if spec_name is None:
        # The spec can be placed in the same folder if used as a standalone package
        spec_name = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'vtk.xml')
        if not os.path.exists(spec_name):
            return
    specs = SpecList.read_from_xml(spec_name, ClassSpec)
    for spec in specs.module_specs:
        globals()[spec.module_name] = gen_instance_factory(spec)


# Initialize if possible
initialize()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core import debug
import os
import vistrails.core.configuration
from vistrails.core import system
import shutil
import urllib2
import tempfile
import os.path

##############################################################################
# The local codepaths have '.' replaced by '_' because '.' in the names
# of directories that will be used as python libraries confuses the Python
# runtime.

##############################################################################

class PackageRepository(object):

    def __init__(self):
        upd = system.get_vistrails_directory('userPackageDir')
        if upd is not None:
            self._upd = upd
        else:
            conf = vistrails.core.configuration.get_vistrails_configuration()
            self._upd = os.path.join(conf.get('dotVistrails'), '.userpackages')

    def create_main_directory(self, codepath):
        debug.log("Makedir '%s'" % (os.path.join(self._upd, codepath)))
        os.mkdir(os.path.join(self._upd, codepath))

    ##########################################################################

    def create_directory(self, codepath, filename):
        r = os.path.join(self._upd, codepath, filename)
        debug.log("Makedir '%s'" % r)
        os.mkdir(r)

    def copy_file(self, codepath, filename, local_name):
        r = os.path.join(self._upd,
                         codepath, filename)
        debug.log("Copyfile '%s' to '%s'" % (local_name, r))
        shutil.copyfile(local_name, r)

    ##########################################################################
    # implement these

    # def find_package(self, identifier):
    #     pass

    # def install_package(self, identifier):
    #     pass

    ##########################################################################

    class InvalidPackage(Exception):
        pass

##############################################################################

class LocalPackageRepository(PackageRepository):

    def __init__(self, repository_path):
        PackageRepository.__init__(self)
        self._path = repository_path

    def find_package(self, identifier):
        lst = set(os.listdir(self._path))
        codepath = identifier.replace('.', '_')
        if codepath in lst:
            return codepath

    def install_package(self, codepath):
        debug.log("package found!")
        # read manifest
        try:
            f = open(os.path.join(self._path, codepath, 'MANIFEST'))
        except IOError, e:
            raise PackageRepository.InvalidPackage("Package is missing manifest.")
        # create directory
        self.create_main_directory(codepath)
        for l in f:
            l = l[:-1]
            if len(l) == 0:
                continue
            file_type = l[0]
            l = l[2:]
            if file_type == 'D':
                self.create_directory(codepath, l)
            elif file_type == 'F':
                self.copy_file(codepath, l, os.path.join(self._path, codepath, l))

##############################################################################

class HTTPPackageRepository(PackageRepository):

    def __init__(self, repository_url):
        PackageRepository.__init__(self)
        self._path = repository_url

    def find_package(self, identifier):
        identifier = identifier.replace('.', '_')
        package_url = self._path + '/' + identifier
        try:
            f = urllib2.urlopen(package_url)
            return identifier
        except urllib2.HTTPError:
            return None

    def install_package(self, codepath):
        debug.log("package found!")
        # read manifest
        try:
            f = urllib2.urlopen(self._path + '/' + codepath + '/MANIFEST')
        except urllib2.HTTPError, e:
            raise PackageRepository.InvalidPackage("Package is missing manifest.")
        self.create_main_directory(codepath)
        for l in f:
            l = l[:-1]
            if len(l) == 0:
                continue
            file_type = l[0]
            l = l[2:]
            if file_type == 'D':
                self.create_directory(codepath, l)
            elif file_type == 'F':
                fd, name = tempfile.mkstemp()
                os.close(fd)
                try:
                    fout = open(name, 'w')
                    fin = urllib2.urlopen(self._path + '/' + codepath + '/' + l)
                    fout.write(fin.read()) # There should be a better way to do this
                    fin.close()
                    fout.close()
                    self.copy_file(codepath, l, name)
                finally:
                    os.unlink(name)

##############################################################################

_repository = None
def get_repository():
    global _repository
    if _repository:
        return _repository
    import vistrails.core.configuration
    conf = vistrails.core.configuration.get_vistrails_configuration()
    if conf.check('repositoryHTTPURL'):
        _repository = HTTPPackageRepository(conf.repositoryHTTPURL)
        debug.log("Using HTTP Package Repository @ %s" % conf.repositoryHTTPURL)
    elif conf.check('repositoryLocalPath'):
        _repository = LocalPackageRepository(conf.repositoryLocalPath)
        debug.log("Using Local Repository @ %s" % conf.repositoryLocalPath)
    else:
        _repository = None
    return _repository
